﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using BenLincoln.DebugSymbols.SymDump;

namespace BenLincoln.DebugSymbols
{
    public class DupeCounter
    {
        protected Hashtable _Counts;

        public DupeCounter()
        {
            _Counts = new Hashtable();
        }

        public void HandleName(string name)
        {
            if (_Counts.ContainsKey(name))
            {
                _Counts[name] = ((int)_Counts[name]) + 1;
            }
            else
            {
                _Counts.Add(name, 1);
            }
        }

        public List<string> GetDuplicateNames()
        {
            List<string> result = new List<string>();
            foreach (string k in _Counts.Keys)
            {
                int count = (int)_Counts[k];
                if (count > 1)
                {
                    result.Add(k);
                }
            }
            return result;
        }

        public string GetDupeList()
        {
            StringBuilder builder = new StringBuilder();
            List<string> keyNames = new List<string>();
            foreach (string k in _Counts.Keys)
            {
                keyNames.Add(k);
            }
            keyNames.Sort();
            foreach (string k in keyNames)
            {
                int count = (int)_Counts[k];
                if (count > 1)
                {
                    builder.Append(string.Format("\t{0} - {1} definitions with this name{2}", k, count, Environment.NewLine));
                }
            }

            return builder.ToString();
        }

        public string GetResultOutput(string objectType)
        {
            StringBuilder builder = new StringBuilder();
            string result = GetDupeList();
            if (result != "")
            {
                builder.Append(string.Format("Warning: at least one {0} definition with a duplicate name was detected. You will need to manually resolve any conflicts by renaming the additional definitions in the JSON file, as well as any references to them in the same file, then re-running CreateSkeleton.", objectType));
                builder.Append(Environment.NewLine);
                builder.Append(result);
                builder.Append(Environment.NewLine);
            }
            return builder.ToString();
        }
    }

    public class SourceCodeGenerationResult
    {
        public string SourceCode = "";
        public StringBuilder ManualChangesBuilder = new StringBuilder();
    }

    public class MonolithicHeaderGenerationResult : SourceCodeGenerationResult
    {
        public List<string> DuplicateEnumNames = new List<string>();
        public List<string> DuplicateStructNames = new List<string>();
        public List<string> DuplicateUnionNames = new List<string>();

        public List<string> EnumStructDupeNames = new List<string>();
        public List<string> EnumUnionDupeNames = new List<string>();
        public List<string> StructUnionDupeNames = new List<string>();
    }

    public class SourceCodeGenerator
    {

        public static MonolithicHeaderGenerationResult GenerateMonolithicHeader(SymDumpSymbolCollection symbolCollection, StringBuilder manualChangesBuilder, List<string> functionPointerTypeDefs, List<string> addedEnums, List<string> addedStructs, List<string> addedUnions)
        {
            // create monolithic header file
            MonolithicHeaderGenerationResult result = new MonolithicHeaderGenerationResult();
            StringBuilder mHeaderFileBuilder = new StringBuilder();

            mHeaderFileBuilder.Append("/* Begin enums */");
            mHeaderFileBuilder.Append(Environment.NewLine);

            DupeCounter dc = new DupeCounter();
            foreach (CEnum e in symbolCollection.enums)
            {
                dc.HandleName(e.name);
                mHeaderFileBuilder.Append(Environment.NewLine);
                // mHeaderFileBuilder.Append(String.Format("enum {0} : {1}", e.name, e.primitive_type));
                mHeaderFileBuilder.Append(String.Format("enum {0} // Hashcode: {1}", e.name, e.GetHashCodeAsString()));
                mHeaderFileBuilder.Append(Environment.NewLine);
                mHeaderFileBuilder.Append("{");
                mHeaderFileBuilder.Append(Environment.NewLine);
                int valueNum = 0;
                foreach (string valueName in e.members.Keys)
                {
                    mHeaderFileBuilder.Append(String.Format("\t{0} = {1}", valueName, e.members[valueName]));
                    //valueNum++;
                    //if (valueNum < e.members.Count)
                    //{
                    mHeaderFileBuilder.Append(",");
                    //}
                    mHeaderFileBuilder.Append(Environment.NewLine);
                }
                mHeaderFileBuilder.Append("};");
                mHeaderFileBuilder.Append(Environment.NewLine);
            }
            manualChangesBuilder.Append(dc.GetResultOutput("enum"));
            result.DuplicateEnumNames = dc.GetDuplicateNames();
            dc = new DupeCounter();

            mHeaderFileBuilder.Append("/* End enums */");
            mHeaderFileBuilder.Append(Environment.NewLine);
            mHeaderFileBuilder.Append(Environment.NewLine);


            mHeaderFileBuilder.Append("/* Begin structs */");
            mHeaderFileBuilder.Append(Environment.NewLine);

            foreach (CStruct s in symbolCollection.structs)
            {
                dc.HandleName(s.name);
                mHeaderFileBuilder.Append(Environment.NewLine);
                mHeaderFileBuilder.Append(String.Format("struct {0} // hashcode: {1}", s.name, s.GetHashCodeAsString()));
                mHeaderFileBuilder.Append(Environment.NewLine);
                mHeaderFileBuilder.Append("{");
                mHeaderFileBuilder.Append(Environment.NewLine);
                foreach (CStructMember mem in s.members)
                {
                    // mHeaderFileBuilder.Append(String.Format("\t{0}", mem));
                    mHeaderFileBuilder.Append("\t");
                    mHeaderFileBuilder.Append(mem.struct_member_signature);
                    mHeaderFileBuilder.Append(string.Format(" //  hashcode: {0}, parent name: '{1}', parent hashcode: {2}", mem.GetHashCodeAsString(), mem.parent_name, CElement.GetHashCodeAsString(mem.parent_hashcode)));
                    mHeaderFileBuilder.Append(Environment.NewLine);
                }
                mHeaderFileBuilder.Append("};");
                mHeaderFileBuilder.Append(Environment.NewLine);
                //mHeaderFileBuilder.Append(String.Format("typedef struct {0} {1};", s.name, s.name));
                //mHeaderFileBuilder.Append(Environment.NewLine);
            }
            manualChangesBuilder.Append(dc.GetResultOutput("struct"));
            result.DuplicateStructNames = dc.GetDuplicateNames();
            dc = new DupeCounter();

            mHeaderFileBuilder.Append("/* End structs */");
            mHeaderFileBuilder.Append(Environment.NewLine);
            mHeaderFileBuilder.Append(Environment.NewLine);

            mHeaderFileBuilder.Append("/* Begin unions */");
            mHeaderFileBuilder.Append(Environment.NewLine);

            foreach (CUnion u in symbolCollection.unions)
            {
                dc.HandleName(u.name);
                mHeaderFileBuilder.Append(Environment.NewLine);
                mHeaderFileBuilder.Append(String.Format("union {0} // Hashcode: {1}", u.name, u.GetHashCodeAsString()));
                mHeaderFileBuilder.Append(Environment.NewLine);
                mHeaderFileBuilder.Append("{");
                mHeaderFileBuilder.Append(Environment.NewLine);
                foreach (CStructMember mem in u.members)
                {
                    //mHeaderFileBuilder.Append(String.Format("\t{0}", mem));
                    //mHeaderFileBuilder.Append(Environment.NewLine);
                    mHeaderFileBuilder.Append("\t");
                    mHeaderFileBuilder.Append(mem.struct_member_signature);
                    mHeaderFileBuilder.Append(Environment.NewLine);
                }
                mHeaderFileBuilder.Append("};");
                mHeaderFileBuilder.Append(Environment.NewLine);
                //mHeaderFileBuilder.Append(String.Format("typedef union {0} {1};", u.name, u.name));
                //mHeaderFileBuilder.Append(Environment.NewLine);
            }

            manualChangesBuilder.Append(dc.GetResultOutput("union"));
            result.DuplicateUnionNames = dc.GetDuplicateNames();
            dc = new DupeCounter();

            mHeaderFileBuilder.Append("/* End unions */");
            mHeaderFileBuilder.Append(Environment.NewLine);
            mHeaderFileBuilder.Append(Environment.NewLine);


            mHeaderFileBuilder.Append("/* Begin typedefs */");
            mHeaderFileBuilder.Append(Environment.NewLine);

            foreach (CTypeDef t in symbolCollection.typedefs)
            {
                mHeaderFileBuilder.Append(Environment.NewLine);
                mHeaderFileBuilder.Append(String.Format("typedef {0};", t.typedef_signature));
                mHeaderFileBuilder.Append(Environment.NewLine);
            }

            if (functionPointerTypeDefs.Count > 0)
            {
                mHeaderFileBuilder.Append(Environment.NewLine);
                mHeaderFileBuilder.Append("/* Begin autogenerated function pointer typedefs */");
                mHeaderFileBuilder.Append(Environment.NewLine);

                functionPointerTypeDefs.Sort();

                foreach (string fptd in functionPointerTypeDefs)
                {
                    mHeaderFileBuilder.Append(Environment.NewLine);
                    mHeaderFileBuilder.Append(fptd);
                    mHeaderFileBuilder.Append(Environment.NewLine);
                }
            }

            mHeaderFileBuilder.Append(Environment.NewLine);
            mHeaderFileBuilder.Append("/* End autogenerated function pointer typedefs */");
            mHeaderFileBuilder.Append(Environment.NewLine);
            mHeaderFileBuilder.Append(Environment.NewLine);

            mHeaderFileBuilder.Append("/* End typedefs */");
            mHeaderFileBuilder.Append(Environment.NewLine);
            mHeaderFileBuilder.Append(Environment.NewLine);


            mHeaderFileBuilder.Append("/* Begin external declarations */");
            mHeaderFileBuilder.Append(Environment.NewLine);

            foreach (CExternalDeclaration e in symbolCollection.external_declarations)
            {
                mHeaderFileBuilder.Append(Environment.NewLine);
                mHeaderFileBuilder.Append(e.declaration);
                mHeaderFileBuilder.Append(Environment.NewLine);
            }

            mHeaderFileBuilder.Append("/* End external declarations */");
            mHeaderFileBuilder.Append(Environment.NewLine);
            mHeaderFileBuilder.Append(Environment.NewLine);

            foreach (string enumName in addedEnums)
            {
                foreach (string structName in addedStructs)
                {
                    if (enumName == structName)
                    {
                        manualChangesBuilder.Append(string.Format("The name '{0}' is used to identify both an enum and a struct. You should manually rename one or both in the JSON file to resolve the conflict, then re-run CreateSkeleton.", enumName));
                        manualChangesBuilder.Append(Environment.NewLine);
                        result.EnumStructDupeNames.Add(enumName);
                    }
                }
                foreach (string unionName in addedUnions)
                {
                    if (enumName == unionName)
                    {
                        manualChangesBuilder.Append(string.Format("The name '{0}' is used to identify both an enum and a union. You should manually rename one or both in the JSON file to resolve the conflict, then re-run CreateSkeleton.", enumName));
                        manualChangesBuilder.Append(Environment.NewLine);
                        result.EnumUnionDupeNames.Add(enumName);
                    }
                }
            }
            foreach (string structName in addedStructs)
            {
                foreach (string unionName in addedUnions)
                {
                    if (structName == unionName)
                    {
                        manualChangesBuilder.Append(string.Format("The name '{0}' is used to identify both an struct and a union. You should manually rename one or both in the JSON file to resolve the conflict, then re-run CreateSkeleton.", structName));
                        manualChangesBuilder.Append(Environment.NewLine);
                        result.StructUnionDupeNames.Add(structName);
                    }
                }
            }

            result.ManualChangesBuilder = manualChangesBuilder;
            result.SourceCode = mHeaderFileBuilder.ToString();

            return result;
        }
    }
}
