﻿// CreateSkeleton
// Creates a skeleton PlayStation dev project based on the output of SymDumpTE
// or another source of the appropriate JSON data
// Copyright 2019 Ben Lincoln
// https://www.beneaththewaves.net/

// This file is part of CreateSkeleton.

// CreateSkeleton is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// CreateSkeleton is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with CreateSkeleton (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using Newtonsoft.Json;
using BenLincoln.DebugSymbols;
using BenLincoln.DebugSymbols.SymDump;

namespace CreateSkeleton
{

    class Program
    {
        protected const string ProgramName = "CreateSkeleton";
        protected static string Copyright = "Copyright 2019 Ben Lincoln";
        protected const string InfoURL = "https://www.beneaththewaves.net/Software/";
        protected const int MaxFunctionDefinitionsPerGhidraScriptSubroutine = 100;
        protected const int MaxLabelDefinitionsPerGhidraScriptSubroutine = 50;
        protected const int MaxDataTypeImportsPerGhidraScriptSubroutine = 200;
        protected const string ProjectNameValidationRegexString = @"^[A-Za-z0-9_]{1,16}$";
        protected const string ProjectNameExplanation = "alphanumeric characters and underscore (_)";
        //protected const uint MinAddressForLabels = 0x80000000;
        protected const string GhidraFilesDirectory = "ghidra_files";
        protected const string FunctionPointerTypeDefPrefix = "TDRFuncPtr_";

        // labels that overlap with function names, etc. and cause problems
        protected static List<string> GetBlacklistedLabels()
        {
            List<string> result = new List<string>();
            result.Add("_bss_obj");
            result.Add("_bss_objend");
            result.Add("_bss_org");
            result.Add("_bss_orgend");
            result.Add("_bss_size");

            result.Add("_data_obj");
            result.Add("_data_objend");
            result.Add("_data_org");
            result.Add("_data_orgend");
            result.Add("_data_size");

            result.Add("_rdata_obj");
            result.Add("_rdata_objend");
            result.Add("_rdata_org");
            result.Add("_rdata_orgend");
            result.Add("_rdata_size");

            result.Add("_sbss_obj");
            result.Add("_sbss_objend");
            result.Add("_sbss_org");
            result.Add("_sbss_orgend");
            result.Add("_sbss_size");

            result.Add("_sdata_obj");
            result.Add("_sdata_objend");
            result.Add("_sdata_org");
            result.Add("_sdata_orgend");
            result.Add("_sdata_size");

            result.Add("_text_obj");
            result.Add("_text_objend");
            result.Add("_text_org");
            result.Add("_text_orgend");
            result.Add("_text_size");

            result.Add("__bss_obj");
            result.Add("__bss_objend");
            result.Add("__bss_org");
            result.Add("__bss_orgend");
            result.Add("__bss_size");

            result.Add("__data_obj");
            result.Add("__data_objend");
            result.Add("__data_org");
            result.Add("__data_orgend");
            result.Add("__data_size");

            result.Add("__rdata_obj");
            result.Add("__rdata_objend");
            result.Add("__rdata_org");
            result.Add("__rdata_orgend");
            result.Add("__rdata_size");

            result.Add("__sbss_obj");
            result.Add("__sbss_objend");
            result.Add("__sbss_org");
            result.Add("__sbss_orgend");
            result.Add("__sbss_size");

            result.Add("__sdata_obj");
            result.Add("__sdata_objend");
            result.Add("__sdata_org");
            result.Add("__sdata_orgend");
            result.Add("__sdata_size");

            result.Add("__text_obj");
            result.Add("__text_objend");
            result.Add("__text_org");
            result.Add("__text_orgend");
            result.Add("__text_size");
            return result;
        }

        static void Main(string[] args)
        {
            if ((args.Length < 5))
            {
                DisplayHelpMessage();
                Environment.Exit(1);
            }

            string projectName = "";
            string baseOutputDirectoryPath = "";
            string inputFilePath = "";
            bool forceOverwrite = false;
            bool ignoreLabels = false;
            bool ignoreNonASCIILabels = false;
            bool convertExternalDefinitionsToLabels = false;
            bool addStackPointerOffset = false;
            bool assignReturnAddressStorage = false;
            bool assignParameterStorage = false;
            bool createPlayStationMemory = false;
            bool assumeSNGPBase = false;

            ArrayList arguments = new ArrayList();
            for (int i = 0; i < args.Length; i++)
            {
                // if the argument is one with a second component, catch that now
                string nextArg = "";
                switch (args[i])
                {
                    case "--name":
                        nextArg = args[i + 1];
                        projectName = nextArg;
                        Regex pnRex = new Regex(ProjectNameValidationRegexString);
                        if (!pnRex.IsMatch(projectName))
                        {
                            Console.WriteLine(String.Format("The specified project name is invalid. The name must consist only of {0}.", ProjectNameExplanation));
                            Environment.Exit(1);
                        }
                        i++;
                        break;
                    case "--output":
                        nextArg = args[i + 1];
                        baseOutputDirectoryPath = Path.GetFullPath(nextArg);
                        i++;
                        break;
                    default:
                        arguments.Add(args[i]);
                        break;
                }
            }

            ArrayList positionalArgs = new ArrayList();

            foreach (string arg in arguments)
            {
                if ((arg == "-?") || (arg == "-h") || (arg == "--help"))
                {
                    DisplayHelpMessage();
                    Environment.Exit(0);
                }
                bool handled = false;
                if (arg == "--overwrite")
                {
                    forceOverwrite = true;
                    handled = true;
                }
                if (arg == "--ignore-labels")
                {
                    ignoreLabels = true;
                    handled = true;
                }
                if (arg == "--ignore-non-ascii-labels")
                {
                    ignoreNonASCIILabels = true;
                    handled = true;
                }
                if (arg == "--externs-to-labels")
                {
                    convertExternalDefinitionsToLabels = true;
                    handled = true;
                }
                if (arg == "--add-stack-pointer-offset")
                {
                    addStackPointerOffset = true;
                    handled = true;
                }
                if (arg == "--assign-ra-storage")
                {
                    assignReturnAddressStorage = true;
                    handled = true;
                }
                if (arg == "--assign-parameter-storage")
                {
                    assignParameterStorage = true;
                    handled = true;
                }
                if (arg == "--create-playstation-memory")
                {
                    createPlayStationMemory = true;
                    handled = true;
                }
                if (arg == "--assume-sn-gp-base")
                {
                    assumeSNGPBase = true;
                    handled = true;
                }
                // right now, all positional arguments are file paths
                if (!handled)
                {
                    positionalArgs.Add(arg);
                }
            }

            if (positionalArgs.Count > 0)
            {
                inputFilePath = (string)positionalArgs[0];
            }

            if (inputFilePath == "")
            {
                ExitWithError("No input SymDump file path was specified.", 1);
            }

            if (!File.Exists(inputFilePath))
            {
                ExitWithError("The specified input SymDump file does not exist.", 1);
            }

            if (baseOutputDirectoryPath == "")
            {
                ExitWithError("No output directory was specified.", 1);
            }
            if (forceOverwrite)
            {
                if (Directory.Exists(baseOutputDirectoryPath))
                {
                    try
                    {
                        Directory.Delete(baseOutputDirectoryPath, true);
                    }
                    catch (Exception ex)
                    {
                        ExitWithError(String.Format("The output directory '{0}' already exists, and could not be deleted: {1}", baseOutputDirectoryPath, ex.Message), 2);
                    }
                }
            }
            /*
            else
            {
                if (Directory.Exists(baseOutputDirectoryPath))
                {
                    ExitWithError("The specified output directory already exists. Use the --overwrite option to replace it.", 1);
                }
            }
            */

            // create base output directory structure
            string[] outputDirs = new string[] { "source-stubs", "source-decompiled", GhidraFilesDirectory };
            foreach (string od in outputDirs)
            {
                string fullPath = Path.Combine(baseOutputDirectoryPath, od);
                try
                {
                    Directory.CreateDirectory(fullPath);
                }
                catch (Exception ex)
                {
                    ExitWithError(String.Format("Could not create the output directory '{0}': {1}", fullPath, ex.Message), 2);
                }
            }

            string monolithicHeaderFilePath = Path.Combine(baseOutputDirectoryPath, "source-stubs", String.Format("{0}.H", projectName));
            string ghidraScriptPath = Path.Combine(baseOutputDirectoryPath, GhidraFilesDirectory, String.Format("{0}TDRDefineFunctions.java", projectName));
            string ghidraExportScriptPath = Path.Combine(baseOutputDirectoryPath, GhidraFilesDirectory, String.Format("{0}TDRDecompile.java", projectName));
            string ghidraXMLFilePath = Path.Combine(baseOutputDirectoryPath, GhidraFilesDirectory, String.Format("{0}.xml", projectName));
            string manualChangesPath = Path.Combine(baseOutputDirectoryPath, String.Format("README-{0}-CreateSkeleton-Manual_Changes_Required.txt", projectName));
            string ghidraDecompiledCCodeFilePath = Path.Combine(baseOutputDirectoryPath, String.Format("{0}.C", projectName));


            // get SymDumpTE output
            string symDumpRaw = "";
            try
            {
                symDumpRaw = GetFileAsString(inputFilePath);
            }
            catch (Exception ex)
            {
                ExitWithError(String.Format("Could not read the input file '{0}': {1}", inputFilePath, ex.Message), 2);
            }

            StringBuilder manualChangesBuilder = new StringBuilder();

            // fix names that don't work in Ghidra
            Regex dotRenameRex = new Regex("(?<itemName>(union|struct|enum)( | [^ ]+ \\*)\\.[^ ;\"$]+)[ ;\"$]");
            MatchCollection drMatches = dotRenameRex.Matches(symDumpRaw);
            Dictionary<string, int> dotNames = new Dictionary<string, int>();
            foreach (Match m in drMatches)
            {
                if (m.Groups["itemName"] != null)
                {
                    string matchText = m.Groups["itemName"].Value;
                    if (!dotNames.ContainsKey(matchText))
                    {
                        dotNames.Add(matchText, 1);
                    }
                    else
                    {
                        dotNames[matchText]++;
                    }
                }
            }

            if (dotNames.Count > 0)
            {
                manualChangesBuilder.Append("The input JSON file contains at least one C element with a name that starts with a '.'. This naming convention is incompatible with Ghidra, so it will be replaced by a leading underscore. You should consider renaming these elements in the source JSON file, and will need to do so if any of them have name conflicts listed later in this file.");
                manualChangesBuilder.Append(Environment.NewLine);
                List<string> keyNames = dotNames.Keys.ToList();
                keyNames.Sort();
                manualChangesBuilder.Append("Elements which will be renamed:");
                manualChangesBuilder.Append(Environment.NewLine);
                foreach (string dotName in keyNames)
                {
                    if (dotNames[dotName] > 1)
                    {
                        manualChangesBuilder.Append(string.Format("\t{0} ({1} occurrences)", dotName, dotNames[dotName]));
                    }
                    else
                    {
                        manualChangesBuilder.Append(string.Format("\t{0}", dotName));
                    }
                    manualChangesBuilder.Append(Environment.NewLine);
                }
                manualChangesBuilder.Append(Environment.NewLine);
            }

            symDumpRaw = symDumpRaw.Replace("union .", "union _");
            symDumpRaw = symDumpRaw.Replace("struct .", "struct _");
            symDumpRaw = symDumpRaw.Replace("enum .", "enum _");
            symDumpRaw = symDumpRaw.Replace("\"name\": \".", "\"name\": \"_");
            foreach (string matchText in dotNames.Keys)
            {
                string replaceText = matchText.Replace(".", "_");
                Console.WriteLine(string.Format("Debug: replacing '{0}' with '{1}', {2} occurrences in JSON file", matchText, replaceText, dotNames[matchText]));
                symDumpRaw = symDumpRaw.Replace(matchText, replaceText);
            }

            SymDumpSymbolCollection symbolCollection = JsonConvert.DeserializeObject<SymDumpSymbolCollection>(symDumpRaw);

            // create Ghidra function-definition script
            string ghidraFunctionDefinitionScriptTemplate = GetEmbeddedTextFile("CreateSkeleton.GhidraScript-Template-DefineFunctions.txt");

            string ghidraDefineFunctionsSubroutineTemplate = GetEmbeddedTextFile("CreateSkeleton.GhidraScript-Template-DefineFunctionsSubroutine.txt");

            string ghidraDefineLabelsSubroutineTemplate = GetEmbeddedTextFile("CreateSkeleton.GhidraScript-Template-CreateLabelsSubroutine.txt");

            string ghidraImportDataTypeSubroutineTemplate = GetEmbeddedTextFile("CreateSkeleton.GhidraScript-Template-ImportDataTypesSubroutine.txt");

            string ghidraCustomDecompileExportTemplate = GetEmbeddedTextFile("CreateSkeleton.GhidraScript-Template-TDRDecompile.txt");

            ghidraCustomDecompileExportTemplate = ghidraCustomDecompileExportTemplate.Replace("%DECOMPILED_C_CODE_FILE%", ghidraDecompiledCCodeFilePath.Replace("\\", "\\\\"));

            ghidraCustomDecompileExportTemplate = ghidraCustomDecompileExportTemplate.Replace("%PROJECT_NAME%", projectName);

            WriteToOutput(ghidraExportScriptPath, ghidraCustomDecompileExportTemplate, FileMode.Create);

            // set register values
            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%ENTRYPOINT_ADDRESS%", string.Format("0x{0:X8}", symbolCollection.program_counter));
            uint stackPointerValue = symbolCollection.stack_pointer_base;
            uint snGlobalPointerValue = 0;
            CLabel snGLabel = symbolCollection.GetLabelByName("__SN_GP_BASE");
            if (snGLabel != null)
            {
                snGlobalPointerValue = snGLabel.offset;
                Console.WriteLine(string.Format("Info: found __SN_GP_BASE (global pointer value) 0x{0:X8} in the symbol data.", snGlobalPointerValue));
                if (assumeSNGPBase)
                {
                    Console.WriteLine(string.Format("Info: will request that Ghidra explicitly assume that $gp is set to 0x{0:X8} at the beginning of each function except 'entry'.", snGlobalPointerValue));
                }
                else
                {
                    Console.WriteLine("Warning: The __SN_GP_BASE value will be ignored. It is highly recommended that you re-run CreateSkeleton.exe with the --assume-sn-gp-base flag unless doing so is resulting in bad data.");
                }
            }
            else
            {
                Console.WriteLine("Warning: Did not find an __SN_GP_BASE value in the debug symbol data.");
            }

            if (addStackPointerOffset)
            {
                stackPointerValue += symbolCollection.stack_pointer_size_or_offset;
                if (symbolCollection.stack_pointer_size_or_offset != 0)
                {
                    Console.WriteLine(string.Format("Info: adding the non-zero stack pointer offset value 0x{0:X8} to the stack pointer value.", symbolCollection.stack_pointer_size_or_offset));
                }

            }
            else
            {
                if (symbolCollection.stack_pointer_size_or_offset != 0)
                {
                    Console.WriteLine(string.Format("Warning: the stack pointer size/offset value for this binary is not zero (0x{0:X8}). If you notice bad decompilation results, consider re-running CreateSkeleton.exe with the --add-stack-pointer-offset flag.", symbolCollection.stack_pointer_size_or_offset));
                }
            }
            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%STACK_POINTER_VALUE%", string.Format("0x{0:X8}", stackPointerValue));
            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%GLOBAL_POINTER_VALUE%", string.Format("0x{0:X8}", symbolCollection.global_pointer));


            StringBuilder memoryMapBuilder = new StringBuilder();

            long lowerCodeBound = 0xFFFFFFFF;
            long upperCodeBound = 0x00000000;
            if (symbolCollection.memory_segments.Count > 0)
            {
                memoryMapBuilder.Append("\t\tMemory mem = currentProgram.getMemory();");
                memoryMapBuilder.Append(Environment.NewLine);
                foreach (MemorySegment ms in symbolCollection.memory_segments)
                {
                    memoryMapBuilder.Append(string.Format("\t\t// segment {0} - {1}: 0x{2:X8}-0x{3:X8} (length: 0x{4:X8})", ms.segment_number, ms.name, ms.offset, ms.offset + ms.length, ms.length));
                    memoryMapBuilder.Append(Environment.NewLine);

                    bool includeMemory = true;

                    if (ms.segment_number > 1)
                    {
                        if (createPlayStationMemory)
                        {
                            includeMemory = false;
                        }
                        else
                        {
                            memoryMapBuilder.Append(string.Format("\t\tMemoryBlock mb{0} = mem.createUninitializedBlock(\"{1}\", toAddr(0x{2:X8}), 0x{3:X8}, false);", ms.segment_number, ms.name, ms.offset, ms.length));
                            memoryMapBuilder.Append(Environment.NewLine);
                            memoryMapBuilder.Append(string.Format("\t\tmb{0}.setRead(true);", ms.segment_number));
                            memoryMapBuilder.Append(Environment.NewLine);
                            memoryMapBuilder.Append(string.Format("\t\tmb{0}.setWrite(true);", ms.segment_number));
                            memoryMapBuilder.Append(Environment.NewLine);
                            memoryMapBuilder.Append(string.Format("\t\tmb{0}.setExecute(true);", ms.segment_number));
                            memoryMapBuilder.Append(Environment.NewLine);
                        }
                    }

                    if (includeMemory)
                    {
                        if (ms.offset < lowerCodeBound)
                        {
                            lowerCodeBound = ms.offset;
                        }
                        long segmentUpperBound = ms.offset + ms.length;
                        if (segmentUpperBound > upperCodeBound)
                        {
                            upperCodeBound = segmentUpperBound;
                        }
                    }
                }
            }
            if (createPlayStationMemory)
            {
                if (lowerCodeBound > upperCodeBound)
                {
                    Console.WriteLine(string.Format("Error: Text/code segment lower bound was calculated as 0x{0:X8}, but the upper bound was calculated as 0x{1:X8}, which is below the lower bound. Please contact the developer of this tool for guidance. The additional PlayStation memory map data will not be created.", lowerCodeBound, upperCodeBound));
                }
                else
                {
                    memoryMapBuilder.Append("\t\t// create all additional memory map/segment data");
                    memoryMapBuilder.Append(Environment.NewLine);
                    memoryMapBuilder.Append(string.Format("\t\tsetupMemoryMap(currentProgram, 0x{0:X8}L, 0x{1:X8}L);", lowerCodeBound, (upperCodeBound - lowerCodeBound)));
                    memoryMapBuilder.Append(Environment.NewLine);
                }
            }

            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%MEMORY_MAP_POPULATION%", memoryMapBuilder.ToString());

            StringBuilder labelImportBuilder = new StringBuilder();
            StringBuilder labelSubImportBuilder = new StringBuilder();
            StringBuilder labelImportCallBuilder = new StringBuilder();
            int labelNum = 0;
            int labelsInSubCount = 0;
            int labelSubRoutineNumber = 0;
            List<string> blacklistedLabels = GetBlacklistedLabels();

            if (!ignoreLabels)
            {
                int lNum = 0;
                foreach (CLabel l in symbolCollection.labels)
                {
                    bool includeLabel = true;
                    //if (blacklistedLabels.Contains(l.name))
                    //{
                    //    includeLabel = false;
                    //}
                    //if (l.offset < MinAddressForLabels)
                    //{
                    //    includeLabel = false;
                    //}
                    if (ignoreNonASCIILabels)
                    {
                        if (l.name.Trim() == "")
                        {
                            includeLabel = false;
                            Console.WriteLine(string.Format("Warning: Ignoring blank label at 0x{0:X8}", l.offset));
                        }
                        if (includeLabel)
                        {
                            char[] labelArray = l.name.ToCharArray();
                            for (int charNum = 0; charNum < labelArray.Length; charNum++)
                            {
                                if ((labelArray[charNum] < 20) || (labelArray[charNum] > 127))
                                {
                                    includeLabel = false;
                                    Console.WriteLine(string.Format("Warning: Ignoring label with non-ASCII characters at 0x{0:X8}", l.offset));
                                    break;
                                }
                            }
                        }
                    }
                    if (includeLabel)
                    {
                        labelSubImportBuilder.Append(String.Format("\t\tSymbol sym{0} = symbolTable.createLabel(toAddr(0x{1:X8}), \"{2}\", SourceType.USER_DEFINED);", labelNum, l.offset, l.name));
                        labelSubImportBuilder.Append(Environment.NewLine);
                        labelSubImportBuilder.Append(String.Format("\t\tsym{0}.setPrimary();", labelNum));
                        labelSubImportBuilder.Append(Environment.NewLine);
                        labelNum++;
                        labelsInSubCount++;
                    }
                    lNum++;
                    if ((labelsInSubCount > MaxLabelDefinitionsPerGhidraScriptSubroutine) || (lNum >= symbolCollection.labels.Count))
                    {
                        string subRoutineDefinition = ghidraDefineLabelsSubroutineTemplate.Replace("%LABEL_DEFINITION_SUB_CONTENT%", labelSubImportBuilder.ToString());
                        subRoutineDefinition = subRoutineDefinition.Replace("%CREATE_LABELS_SUBROUTINE_NUMBER%", String.Format("{0}", labelSubRoutineNumber));
                        labelImportBuilder.Append(subRoutineDefinition);
                        labelImportCallBuilder.Append(String.Format("\t\tCreateLabels{0}(dataTypeMap);", labelSubRoutineNumber));
                        labelImportCallBuilder.Append(Environment.NewLine);
                        labelSubImportBuilder.Clear();
                        labelsInSubCount = 0;
                        labelSubRoutineNumber++;
                    }                    
                }
            }

            if (convertExternalDefinitionsToLabels)
            {
                int xNum = 0;
                foreach (CExternalDeclaration x in symbolCollection.external_declarations)
                {
                    // create the label for the extern/static
                    string externName = x.GetName();
                    labelSubImportBuilder.Append(String.Format("\t\tSymbol sym{0} = symbolTable.createLabel(toAddr(0x{1:X8}), \"{2}\", SourceType.USER_DEFINED);", labelNum, x.offset, externName));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append(String.Format("\t\tsym{0}.setPrimary();", labelNum));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    // TKTK: set the data type for the bytes of the thing
                    // note: array size 0 in a struct definition means basically "occupies zero space in the struct itself, 
                    //       but the struct may be followed by an arbitrary number of whatever this thing is"
                    //       https://gcc.gnu.org/onlinedocs/gcc/Zero-Length.html
                    // GetDataTypeFromMap(dataTypeMap, "struct AadSeqEvent *")
                    // createData(addr, wordDataType);
                    int arrayCount = x.GetCount();
                    string typeName = x.GetTypeName();
                    labelSubImportBuilder.Append(String.Format("\t\tlong xAddr{0} = {1}L;", labelNum, x.offset));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append(String.Format("\t\tDataType xDataType{0} = GetDataTypeFromMap(dataTypeMap, \"{1}\");", labelNum, typeName));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append(String.Format("\t\tif (xDataType{0} != null);", labelNum));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append("\t\t{");
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append(String.Format("\t\t\tlong typeSize = xDataType{0}.getLength();", labelNum));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    //for (int i = 0; i < arrayCount; i++)
                    //{
                    labelSubImportBuilder.Append(String.Format("\t\t\tfor (int i = 0; i < {0}; i++)", arrayCount));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append("\t\t\t{");
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append(String.Format("\t\t\t\tClearAndCreateData(xAddr{0}, typeSize, xDataType{0});", labelNum));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append(String.Format("\t\t\t\txAddr{0} = xAddr{0} + typeSize;", labelNum));
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelSubImportBuilder.Append("\t\t\t}");
                    labelSubImportBuilder.Append(Environment.NewLine);
                    //}
                    labelSubImportBuilder.Append("\t\t}");
                    labelSubImportBuilder.Append(Environment.NewLine);
                    labelNum++;
                    labelsInSubCount++;
                    xNum++;
                    if ((labelsInSubCount > MaxLabelDefinitionsPerGhidraScriptSubroutine) || (xNum >= (symbolCollection.external_declarations.Count - 1)))
                    {
                        string subRoutineDefinition = ghidraDefineLabelsSubroutineTemplate.Replace("%LABEL_DEFINITION_SUB_CONTENT%", labelSubImportBuilder.ToString());
                        subRoutineDefinition = subRoutineDefinition.Replace("%CREATE_LABELS_SUBROUTINE_NUMBER%", String.Format("{0}", labelSubRoutineNumber));
                        labelImportBuilder.Append(subRoutineDefinition);
                        labelImportCallBuilder.Append(String.Format("\t\tCreateLabels{0}(dataTypeMap);", labelSubRoutineNumber));
                        labelImportCallBuilder.Append(Environment.NewLine);
                        labelSubImportBuilder.Clear();
                        labelsInSubCount = 0;
                        labelSubRoutineNumber++;
                    }
                }
            }
            //// flush any remaining label definitions to the buffer
            //if (labelsInSubCount > 0)
            //{
            //    string subRoutineDefinition = ghidraDefineLabelsSubroutineTemplate.Replace("%LABEL_DEFINITION_SUB_CONTENT%", labelSubImportBuilder.ToString());
            //    subRoutineDefinition = subRoutineDefinition.Replace("%CREATE_LABELS_SUBROUTINE_NUMBER%", String.Format("{0}", labelSubRoutineNumber));
            //    labelImportBuilder.Append(subRoutineDefinition);
            //    labelImportCallBuilder.Append(String.Format("\t\tCreateLabels{0}(dataTypeMap);", labelSubRoutineNumber));
            //    labelImportCallBuilder.Append(Environment.NewLine);
            //    labelSubImportBuilder.Clear();
            //    labelsInSubCount = 0;
            //    labelSubRoutineNumber++;
            //}

            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%LABEL_CREATION_SUBROUTINES%", labelImportBuilder.ToString());

            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%LABEL_CREATION_SUBROUTINE_CALLS%", labelImportCallBuilder.ToString());

            int dtImportSubNum = 0;

            AddDataTypeByNameResult enumAddResult = GhidraScriptGenerator.GetAddDataTypesScriptCode(ghidraImportDataTypeSubroutineTemplate,
                symbolCollection.enums, projectName, "enum", MaxDataTypeImportsPerGhidraScriptSubroutine, dtImportSubNum);
            dtImportSubNum = enumAddResult.NextSubroutineNumber;
            List<string> addedEnums = enumAddResult.AddedNames;
            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%ENUM_IMPORT_SUBROUTINE_CALLS%", enumAddResult.GhidraCodeSubroutineCalls);
            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%ENUM_IMPORT_SUBROUTINES%", enumAddResult.GhidraCodeSubroutines);

            AddDataTypeByNameResult structAddResult = GhidraScriptGenerator.GetAddDataTypesScriptCode(ghidraImportDataTypeSubroutineTemplate,
                symbolCollection.structs, projectName, "struct", MaxDataTypeImportsPerGhidraScriptSubroutine, dtImportSubNum);
            dtImportSubNum = structAddResult.NextSubroutineNumber;
            List<string> addedStructs = structAddResult.AddedNames;
            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%STRUCT_IMPORT_SUBROUTINE_CALLS%", structAddResult.GhidraCodeSubroutineCalls);
            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%STRUCT_IMPORT_SUBROUTINES%", structAddResult.GhidraCodeSubroutines);

            AddDataTypeByNameResult unionAddResult = GhidraScriptGenerator.GetAddDataTypesScriptCode(ghidraImportDataTypeSubroutineTemplate,
                symbolCollection.unions, projectName, "union", MaxDataTypeImportsPerGhidraScriptSubroutine, dtImportSubNum);
            dtImportSubNum = unionAddResult.NextSubroutineNumber;
            List<string> addedUnions = unionAddResult.AddedNames;
            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%UNION_IMPORT_SUBROUTINE_CALLS%", unionAddResult.GhidraCodeSubroutineCalls);
            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%UNION_IMPORT_SUBROUTINES%", unionAddResult.GhidraCodeSubroutines);

            //StringBuilder enumImportBuilder = new StringBuilder();
            //List<string> addedEnums = new List<string>();
            //foreach (CEnum e in symbolCollection.enums)
            //{
            //    if (!addedEnums.Contains(e.name))
            //    {
            //        enumImportBuilder.Append(String.Format("\t\tdataTypeMap = addDataTypeByName(dataTypeMap, \"/{0}.H/{1}\", \"{1}\", \"enum\");", projectName, e.name));
            //        enumImportBuilder.Append(Environment.NewLine);
            //        addedEnums.Add(e.name);
            //    }
            //}
            //ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%ENUM_IMPORTS%", enumImportBuilder.ToString());

            //StringBuilder structImportBuilder = new StringBuilder();
            //List<string> addedStructs = new List<string>();
            //foreach (CStruct s in symbolCollection.structs)
            //{
            //    if (!addedStructs.Contains(s.name))
            //    {
            //        structImportBuilder.Append(String.Format("\t\tdataTypeMap = addDataTypeByName(dataTypeMap, \"/{0}.H/{1}\", \"{1}\", \"struct\");", projectName, s.name));
            //        structImportBuilder.Append(Environment.NewLine);
            //        addedStructs.Add(s.name);
            //    }
            //}
            //ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%STRUCT_IMPORTS%", structImportBuilder.ToString());

            //StringBuilder unionImportBuilder = new StringBuilder();
            //List<string> addedUnions = new List<string>();
            //foreach (CUnion u in symbolCollection.unions)
            //{
            //    if (!addedUnions.Contains(u.name))
            //    {
            //        unionImportBuilder.Append(String.Format("\t\tdataTypeMap = addDataTypeByName(dataTypeMap, \"/{0}.H/{1}\", \"{1}\", \"union\");", projectName, u.name));
            //        unionImportBuilder.Append(Environment.NewLine);
            //        addedUnions.Add(u.name);
            //    }
            //}
            //ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%UNION_IMPORTS%", unionImportBuilder.ToString());

            StringBuilder functionImportBuilder = new StringBuilder();
            StringBuilder functionSubImportBuilder = new StringBuilder();
            StringBuilder functionSubSubImportBuilder = new StringBuilder();
            int functionNum = 0;
            int subRoutineNumber = 0;
            int subroutineFunctionCount = 0;
            StringBuilder fpFunctionImportBuilder = new StringBuilder();
            StringBuilder fpFunctionSubImportBuilder = new StringBuilder();
            StringBuilder fpFunctionSubSubImportBuilder = new StringBuilder();
            int fpFunctionNum = 0;
            int fpSubRoutineNumber = 0;
            int fpSubroutineFunctionCount = 0;
            List<CFunction> functionsWithFunctionPointerParams = new List<CFunction>();
            List<string> functionsReferredToByPointers = new List<string>();
            //Dictionary<string, List<CFunction>> sourceFiles = new Dictionary<string, List<CFunction>>();
            BenLincoln.DebugSymbols.FunctionSourceTracker fsTracker = new BenLincoln.DebugSymbols.FunctionSourceTracker();

            //quick hack to handle function pointer parameters, this should probably be done better, but they should be pretty rare
            List<string> functionPointerTypeDefs = new List<string>();
            List<string> functionPointerFunctionNames = new List<string>();
            List<string> functionPointerNames = new List<string>();

            foreach (CFunction func in symbolCollection.functions)
            {
                try
                {
                    fsTracker.AddOrUpdateEntry(func);
                    if (func.HasFunctionPointerParameters())
                    {
                        manualChangesBuilder.Append(String.Format("Method {0} has a return type and/or parameters which are function pointers. You should manually verify in Ghidra that they were imported correctly, as they can be slightly problematic.", func.name));
                        manualChangesBuilder.Append(Environment.NewLine);
                        manualChangesBuilder.Append(String.Format("\tMethod signature: {0}", func.method_signature));
                        manualChangesBuilder.Append(Environment.NewLine);
                        manualChangesBuilder.Append(String.Format("\tMemory offset: 0x{0:X8}", func.offset));
                        manualChangesBuilder.Append(Environment.NewLine);
                        manualChangesBuilder.Append(String.Format("\tFile: {0}", func.source_file));
                        manualChangesBuilder.Append(Environment.NewLine);
                        manualChangesBuilder.Append(String.Format("\tOriginal line number: {0}", func.source_file_line_number));
                        manualChangesBuilder.Append(Environment.NewLine);
                        manualChangesBuilder.Append(String.Format("\tParameters which are function pointers:"));
                        manualChangesBuilder.Append(Environment.NewLine);
                        if (func.return_value.is_function_pointer)
                        {
                            string fptd = GetFunctionPointerTypeDefFromParameter(func.return_value);
                            // this is a hack, but I can't think of a better approach ATM
                            string newTypeName = string.Format("{0}{1}", FunctionPointerTypeDefPrefix, func.name);
                            string oldTypeName = func.return_value.type_name;
                            func.return_value.type_name = newTypeName;
                            func.method_signature = func.method_signature.Replace(oldTypeName, newTypeName);
                            if (!functionPointerNames.Contains(newTypeName))
                            {
                                functionPointerNames.Add(newTypeName);
                            }
                            fptd = fptd.Replace("*)(", string.Format("*{0})(", newTypeName));
                            int lastSpacePos = fptd.LastIndexOf(" ");
                            if (lastSpacePos > 0)
                            {
                                fptd = fptd.Substring(0, lastSpacePos).Trim() + ";";
                            }

                            if (!functionPointerTypeDefs.Contains(fptd))
                            {
                                functionPointerTypeDefs.Add(fptd);
                            }
                            manualChangesBuilder.Append(String.Format("\t\t(The return type of the method)"));
                            manualChangesBuilder.Append(Environment.NewLine);
                        }
                        int paramNum = 0;
                        foreach (CParameter p in func.parameters)
                        {
                            if (p.is_function_pointer)
                            {
                                string fptd = GetFunctionPointerTypeDefFromParameter(p);

                                string newTypeName = string.Format("{0}{1}{2}{3}", FunctionPointerTypeDefPrefix, func.name, paramNum, p.name);
                                string oldTypeName = p.type_name;
                                p.type_name = newTypeName;
                                string oldTypeNameWithVarName = p.parameter_signature;
                                int paramSigCommentPos = p.parameter_signature.LastIndexOf("/*");
                                string paramSigComment = "";
                                if (paramSigCommentPos > 0)
                                {
                                    paramSigComment = p.parameter_signature.Substring(paramSigCommentPos);
                                    oldTypeNameWithVarName = oldTypeNameWithVarName.Substring(0, paramSigCommentPos - 1).Trim();
                                }
                                string newTypeNameWithVarName = string.Format("{0} {1}", newTypeName, p.name);
                                p.parameter_signature = p.parameter_signature.Replace(oldTypeNameWithVarName, newTypeNameWithVarName);
                                if (!functionPointerNames.Contains(newTypeName))
                                {
                                    functionPointerNames.Add(newTypeName);
                                }
                                fptd = string.Format("typedef {0};", oldTypeName);
                                fptd = fptd.Replace("*)(", string.Format("*{0})(", newTypeName));

                                if (!functionPointerTypeDefs.Contains(fptd))
                                {
                                    functionPointerTypeDefs.Add(fptd);
                                }
                                if (!functionPointerFunctionNames.Contains(p.name))
                                {
                                    functionPointerFunctionNames.Add(p.name);
                                }
                                if (!functionPointerNames.Contains(p.name))
                                {
                                    functionPointerNames.Add(p.name);
                                }

                                manualChangesBuilder.Append(String.Format("\t\t{0}", p.name));
                                manualChangesBuilder.Append(Environment.NewLine);
                            }
                            paramNum++;
                        }
                        manualChangesBuilder.Append(Environment.NewLine);
                        functionsWithFunctionPointerParams.Add(func);
                        GenerateFunctionSectionsResult functionPointerParamsFuncResult = GenerateFunctionSections(func, fpFunctionImportBuilder,
                            fpFunctionSubImportBuilder, fpFunctionSubSubImportBuilder, fpFunctionNum, fpSubRoutineNumber, fpSubroutineFunctionCount,
                            ghidraDefineFunctionsSubroutineTemplate, false, "FunctionPointerFunctions", assignReturnAddressStorage, assignParameterStorage, snGlobalPointerValue, assumeSNGPBase);
                        fpFunctionImportBuilder = functionPointerParamsFuncResult.FunctionImportBuilder;
                        fpFunctionSubImportBuilder = functionPointerParamsFuncResult.FunctionSubImportBuilder;
                        fpFunctionNum = functionPointerParamsFuncResult.FunctionNumber;
                        fpSubRoutineNumber = functionPointerParamsFuncResult.SubRoutineNumber;
                        fpSubroutineFunctionCount = functionPointerParamsFuncResult.SubroutineFunctionCount;
                        fpFunctionNum++;
                    }
                    else
                    {
                        GenerateFunctionSectionsResult noFunctionPointerParamsFuncResult = GenerateFunctionSections(func, functionImportBuilder,
                            functionSubImportBuilder, functionSubSubImportBuilder, functionNum, subRoutineNumber, subroutineFunctionCount,
                            ghidraDefineFunctionsSubroutineTemplate, false, "NormalFunctions", assignReturnAddressStorage, assignParameterStorage, snGlobalPointerValue, assumeSNGPBase);
                        functionImportBuilder = noFunctionPointerParamsFuncResult.FunctionImportBuilder;
                        functionSubImportBuilder = noFunctionPointerParamsFuncResult.FunctionSubImportBuilder;
                        functionNum = noFunctionPointerParamsFuncResult.FunctionNumber;
                        subRoutineNumber = noFunctionPointerParamsFuncResult.SubRoutineNumber;
                        subroutineFunctionCount = noFunctionPointerParamsFuncResult.SubroutineFunctionCount;
                        functionNum++;
                    }
                    //}
                }
                // disabled for debugging by using an exception that's never thrown
                catch (OutOfMemoryException ex)
                {
                    Console.WriteLine(String.Format("Error parsing method '{0}', signature '{1}': {2}", func.name, func.method_signature, ex.Message));
                }
            }

            // flush any remaining entries to the StringBuilders
            GenerateFunctionSectionsResult finalFunctionPointerParamsFuncResult = FlushLastEntriesForFunctionSections(fpFunctionImportBuilder,
                           fpFunctionSubImportBuilder, fpFunctionSubSubImportBuilder, fpFunctionNum, fpSubRoutineNumber, fpSubroutineFunctionCount,
                           ghidraDefineFunctionsSubroutineTemplate, "FunctionPointerFunctions");
            fpFunctionImportBuilder = finalFunctionPointerParamsFuncResult.FunctionImportBuilder;
            fpFunctionSubImportBuilder = finalFunctionPointerParamsFuncResult.FunctionSubImportBuilder;
            fpFunctionNum = finalFunctionPointerParamsFuncResult.FunctionNumber;
            fpSubRoutineNumber = finalFunctionPointerParamsFuncResult.SubRoutineNumber;
            fpSubroutineFunctionCount = finalFunctionPointerParamsFuncResult.SubroutineFunctionCount;
            fpFunctionNum++;

            GenerateFunctionSectionsResult finalNoFunctionPointerParamsFuncResult = FlushLastEntriesForFunctionSections(functionImportBuilder,
               functionSubImportBuilder, functionSubSubImportBuilder, functionNum, subRoutineNumber, subroutineFunctionCount,
               ghidraDefineFunctionsSubroutineTemplate, "NormalFunctions");
            functionImportBuilder = finalNoFunctionPointerParamsFuncResult.FunctionImportBuilder;
            functionSubImportBuilder = finalNoFunctionPointerParamsFuncResult.FunctionSubImportBuilder;
            functionNum = finalNoFunctionPointerParamsFuncResult.FunctionNumber;
            subRoutineNumber = finalNoFunctionPointerParamsFuncResult.SubRoutineNumber;
            subroutineFunctionCount = finalNoFunctionPointerParamsFuncResult.SubroutineFunctionCount;
            functionNum++;

            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%FUNCTION_IMPORT_SUBROUTINE_CALLS_1%", functionImportBuilder.ToString());

            functionImportBuilder.Clear();

            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%FUNCTIONS_WITH_FUNCTION_POINTER_PARAMETERS_CALLS%", fpFunctionImportBuilder.ToString());

            fpFunctionImportBuilder.Clear();

            StringBuilder functionPointerImportBuilder = new StringBuilder();
            List<string> addedFunctionPointers = new List<string>();
            foreach (string fp in functionPointerNames)
            {
                if (!addedFunctionPointers.Contains(fp))
                {
                    functionPointerImportBuilder.Append(String.Format("\t\tdataTypeMap = addDataTypeByName(dataTypeMap, \"/{0}.H/{1}\", \"{1}\", \"\");", projectName, fp));
                    functionPointerImportBuilder.Append(Environment.NewLine);
                    addedFunctionPointers.Add(fp);
                }
            }
            //
            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%FUNCTION_POINTER_IMPORTS%", functionPointerImportBuilder.ToString());

            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%FUNCTION_IMPORT_SUBROUTINES%", functionSubImportBuilder.ToString());
            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%FUNCTION_POINTER_FUNCTION_IMPORT_SUBROUTINES%", fpFunctionSubImportBuilder.ToString());

            ghidraFunctionDefinitionScriptTemplate = ghidraFunctionDefinitionScriptTemplate.Replace("%PROJECT_NAME%", projectName);

            WriteToOutput(ghidraScriptPath, ghidraFunctionDefinitionScriptTemplate, FileMode.Create);

            //List<string> duplicateEnumNames = new List<string>();
            //List<string> duplicateStructNames = new List<string>();
            //List<string> duplicateUnionNames = new List<string>();

            // create monolithic header file
            //StringBuilder mHeaderFileBuilder = new StringBuilder();

            //mHeaderFileBuilder.Append("/* Begin enums */");
            //mHeaderFileBuilder.Append(Environment.NewLine);

            //DupeCounter dc = new DupeCounter();
            //foreach (CEnum e in symbolCollection.enums)
            //{
            //    dc.HandleName(e.name);
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //    // mHeaderFileBuilder.Append(String.Format("enum {0} : {1}", e.name, e.primitive_type));
            //    mHeaderFileBuilder.Append(String.Format("enum {0} // Hashcode: {1}", e.name, e.GetHashCodeAsString()));
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //    mHeaderFileBuilder.Append("{");
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //    int valueNum = 0;
            //    foreach (string valueName in e.members.Keys)
            //    {
            //        mHeaderFileBuilder.Append(String.Format("\t{0} = {1}", valueName, e.members[valueName]));
            //        //valueNum++;
            //        //if (valueNum < e.members.Count)
            //        //{
            //        mHeaderFileBuilder.Append(",");
            //        //}
            //        mHeaderFileBuilder.Append(Environment.NewLine);
            //    }
            //    mHeaderFileBuilder.Append("};");
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //}
            //manualChangesBuilder.Append(dc.GetResultOutput("enum"));
            //duplicateEnumNames = dc.GetDuplicateNames();
            //dc = new DupeCounter();

            //mHeaderFileBuilder.Append("/* End enums */");
            //mHeaderFileBuilder.Append(Environment.NewLine);
            //mHeaderFileBuilder.Append(Environment.NewLine);


            //mHeaderFileBuilder.Append("/* Begin structs */");
            //mHeaderFileBuilder.Append(Environment.NewLine);

            //foreach (CStruct s in symbolCollection.structs)
            //{
            //    dc.HandleName(s.name);
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //    mHeaderFileBuilder.Append(String.Format("struct {0} // hashcode: {1}", s.name, s.GetHashCodeAsString()));
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //    mHeaderFileBuilder.Append("{");
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //    foreach (CStructMember mem in s.members)
            //    {
            //        // mHeaderFileBuilder.Append(String.Format("\t{0}", mem));
            //        mHeaderFileBuilder.Append("\t");
            //        mHeaderFileBuilder.Append(mem.struct_member_signature);
            //        mHeaderFileBuilder.Append(string.Format(" //  hashcode: {0}, parent name: '{1}', parent hashcode: {2}", mem.GetHashCodeAsString(), mem.parent_name, CElement.GetHashCodeAsString(mem.parent_hashcode)));
            //        mHeaderFileBuilder.Append(Environment.NewLine);
            //    }
            //    mHeaderFileBuilder.Append("};");
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //    //mHeaderFileBuilder.Append(String.Format("typedef struct {0} {1};", s.name, s.name));
            //    //mHeaderFileBuilder.Append(Environment.NewLine);
            //}
            //manualChangesBuilder.Append(dc.GetResultOutput("struct"));
            //duplicateStructNames = dc.GetDuplicateNames();
            //dc = new DupeCounter();

            //mHeaderFileBuilder.Append("/* End structs */");
            //mHeaderFileBuilder.Append(Environment.NewLine);
            //mHeaderFileBuilder.Append(Environment.NewLine);

            //mHeaderFileBuilder.Append("/* Begin unions */");
            //mHeaderFileBuilder.Append(Environment.NewLine);

            //foreach (CUnion u in symbolCollection.unions)
            //{
            //    dc.HandleName(u.name);
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //    mHeaderFileBuilder.Append(String.Format("union {0} // Hashcode: {1}", u.name, u.GetHashCodeAsString()));
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //    mHeaderFileBuilder.Append("{");
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //    foreach (CStructMember mem in u.members)
            //    {
            //        //mHeaderFileBuilder.Append(String.Format("\t{0}", mem));
            //        //mHeaderFileBuilder.Append(Environment.NewLine);
            //        mHeaderFileBuilder.Append("\t");
            //        mHeaderFileBuilder.Append(mem.struct_member_signature);
            //        mHeaderFileBuilder.Append(Environment.NewLine);
            //    }
            //    mHeaderFileBuilder.Append("};");
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //    //mHeaderFileBuilder.Append(String.Format("typedef union {0} {1};", u.name, u.name));
            //    //mHeaderFileBuilder.Append(Environment.NewLine);
            //}

            //manualChangesBuilder.Append(dc.GetResultOutput("union"));
            //duplicateUnionNames = dc.GetDuplicateNames();
            //dc = new DupeCounter();

            //mHeaderFileBuilder.Append("/* End unions */");
            //mHeaderFileBuilder.Append(Environment.NewLine);
            //mHeaderFileBuilder.Append(Environment.NewLine);


            //mHeaderFileBuilder.Append("/* Begin typedefs */");
            //mHeaderFileBuilder.Append(Environment.NewLine);

            //foreach (CTypeDef t in symbolCollection.typedefs)
            //{
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //    mHeaderFileBuilder.Append(String.Format("typedef {0};", t.typedef_signature));
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //}

            //if (functionPointerTypeDefs.Count > 0)
            //{
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //    mHeaderFileBuilder.Append("/* Begin autogenerated function pointer typedefs */");
            //    mHeaderFileBuilder.Append(Environment.NewLine);

            //    functionPointerTypeDefs.Sort();

            //    foreach (string fptd in functionPointerTypeDefs)
            //    {
            //        mHeaderFileBuilder.Append(Environment.NewLine);
            //        mHeaderFileBuilder.Append(fptd);
            //        mHeaderFileBuilder.Append(Environment.NewLine);
            //    }
            //}

            //mHeaderFileBuilder.Append(Environment.NewLine);
            //mHeaderFileBuilder.Append("/* End autogenerated function pointer typedefs */");
            //mHeaderFileBuilder.Append(Environment.NewLine);
            //mHeaderFileBuilder.Append(Environment.NewLine);

            //mHeaderFileBuilder.Append("/* End typedefs */");
            //mHeaderFileBuilder.Append(Environment.NewLine);
            //mHeaderFileBuilder.Append(Environment.NewLine);


            //mHeaderFileBuilder.Append("/* Begin external declarations */");
            //mHeaderFileBuilder.Append(Environment.NewLine);

            //foreach (CExternalDeclaration e in symbolCollection.external_declarations)
            //{
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //    //mHeaderFileBuilder.Append(String.Format("extern {0}", e.declaration));
            //    mHeaderFileBuilder.Append(e.declaration);
            //    mHeaderFileBuilder.Append(Environment.NewLine);
            //}

            //mHeaderFileBuilder.Append("/* End external declarations */");
            //mHeaderFileBuilder.Append(Environment.NewLine);
            //mHeaderFileBuilder.Append(Environment.NewLine);

            //// check for duplicate names between enums/structs/unions
            //List<string> enumStructDupeNames = new List<string>();
            //List<string> enumUnionDupeNames = new List<string>();
            //List<string> structUnionDupeNames = new List<string>();

            //foreach (string enumName in addedEnums)
            //{
            //    foreach (string structName in addedStructs)
            //    {
            //        if (enumName == structName)
            //        {
            //            manualChangesBuilder.Append(string.Format("The name '{0}' is used to identify both an enum and a struct. You should manually rename one or both in the JSON file to resolve the conflict, then re-run CreateSkeleton.", enumName));
            //            manualChangesBuilder.Append(Environment.NewLine);
            //            enumStructDupeNames.Add(enumName);
            //        }
            //    }
            //    foreach (string unionName in addedUnions)
            //    {
            //        if (enumName == unionName)
            //        {
            //            manualChangesBuilder.Append(string.Format("The name '{0}' is used to identify both an enum and a union. You should manually rename one or both in the JSON file to resolve the conflict, then re-run CreateSkeleton.", enumName));
            //            manualChangesBuilder.Append(Environment.NewLine);
            //            enumUnionDupeNames.Add(enumName);
            //        }
            //    }
            //}
            //foreach (string structName in addedStructs)
            //{
            //    foreach (string unionName in addedUnions)
            //    {
            //        if (structName == unionName)
            //        {
            //            manualChangesBuilder.Append(string.Format("The name '{0}' is used to identify both an struct and a union. You should manually rename one or both in the JSON file to resolve the conflict, then re-run CreateSkeleton.", structName));
            //            manualChangesBuilder.Append(Environment.NewLine);
            //            structUnionDupeNames.Add(structName);
            //        }
            //    }
            //}

            //string monolithicHeaderContent = mHeaderFileBuilder.ToString();

            MonolithicHeaderGenerationResult mhgr = SourceCodeGenerator.GenerateMonolithicHeader(symbolCollection, manualChangesBuilder, functionPointerTypeDefs, addedEnums, addedStructs, addedUnions);
            manualChangesBuilder = mhgr.ManualChangesBuilder;

            WriteToOutput(monolithicHeaderFilePath, mhgr.SourceCode, FileMode.Create);

            // create Ghidra debug symbol XML file even though this doesn't work right now
            WriteToOutput(ghidraXMLFilePath, symbolCollection.ToGhidraXML(String.Format("{0}.SYM", projectName), String.Format("{0}.EXE", projectName)), FileMode.Create);

            // help operator locate duplicate names in the monolithic header
            manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.DuplicateEnumNames, "enum", "Header", monolithicHeaderFilePath, mhgr.SourceCode, "which is defined multiple times");
            manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.DuplicateStructNames, "struct", "Header", monolithicHeaderFilePath, mhgr.SourceCode, "which is defined multiple times");
            manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.DuplicateUnionNames, "union", "Header", monolithicHeaderFilePath, mhgr.SourceCode, "which is defined multiple times");

            manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.EnumStructDupeNames, "enum or struct", "Header", monolithicHeaderFilePath, mhgr.SourceCode, "which is used as a name for both an enum and a struct");
            manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.EnumUnionDupeNames, "enum or union", "Header", monolithicHeaderFilePath, mhgr.SourceCode, "which is used as a name for both an enum and a union");
            manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.StructUnionDupeNames, "struct or union", "Header", monolithicHeaderFilePath, mhgr.SourceCode, "which is used as a name for both a struct and a union");

            // create stubs for individual source files
            foreach (string sourceFileName in fsTracker.SourceFiles.Keys)
            {
                string stubSourceFilePath = sourceFileName;
                // remove any leading drive letter if the code was compiled on Windows
                Regex driveLetterRex = new Regex(@"^[A-Za-z]:");
                if (driveLetterRex.IsMatch(stubSourceFilePath))
                {
                    stubSourceFilePath = stubSourceFilePath.Substring(2);
                }
                // replace forward and backslashes with the path separator for the current platform
                stubSourceFilePath = stubSourceFilePath.Replace("/", Path.DirectorySeparatorChar.ToString());
                stubSourceFilePath = stubSourceFilePath.Replace(@"\", Path.DirectorySeparatorChar.ToString());
                // make path relative to output directory
                while (stubSourceFilePath.StartsWith(Path.DirectorySeparatorChar.ToString()))
                {
                    stubSourceFilePath = stubSourceFilePath.Substring(1);
                }
                string stubTemp = stubSourceFilePath;
                stubSourceFilePath = Path.Combine(baseOutputDirectoryPath, "source-stubs", stubTemp);
                string skeletonSourceFilePath = Path.Combine(baseOutputDirectoryPath, "source-decompiled", stubTemp);

                // get a list of unique line numbers for methods from the file, and sort by that
                List<uint> lineNumbers = new List<uint>();
                foreach (CFunction f in fsTracker.SourceFiles[sourceFileName])
                {
                    if (!lineNumbers.Contains(f.source_file_line_number))
                    {
                        lineNumbers.Add(f.source_file_line_number);
                    }
                }
                lineNumbers.Sort();

                StringBuilder headerFileBuilder = new StringBuilder();
                StringBuilder sourceFileStubBuilder = new StringBuilder();

                string headerFilePath = Path.ChangeExtension(stubSourceFilePath, "H");
                string headerFileName = Path.GetFileName(headerFilePath);
                string headerFileDefString = headerFileName.Replace(".", "_");

                bool headerAndSourceAreSameFile = false;
                if (Path.GetFileName(stubSourceFilePath) == headerFileName)
                {
                    headerAndSourceAreSameFile = true;
                }

                headerFileBuilder.Append(String.Format("#ifndef {0}", headerFileDefString));
                headerFileBuilder.Append(Environment.NewLine);
                headerFileBuilder.Append(String.Format("#define {0}", headerFileDefString));
                headerFileBuilder.Append(Environment.NewLine);
                headerFileBuilder.Append(Environment.NewLine);

                // handle edge case: source file IS header file
                // i.e. if functions are defined in a .H file instead of just declared in it.
                if (!headerAndSourceAreSameFile)
                {
                    sourceFileStubBuilder.Append(String.Format("#include \"{0}.H\"", projectName));
                    sourceFileStubBuilder.Append(Environment.NewLine);
                    sourceFileStubBuilder.Append(String.Format("#include \"{0}\"", headerFileName));
                    sourceFileStubBuilder.Append(Environment.NewLine);
                }
                sourceFileStubBuilder.Append(Environment.NewLine);

                foreach (int lineNumber in lineNumbers)
                {
                    foreach (CFunction f in fsTracker.SourceFiles[sourceFileName])
                    {
                        if (f.source_file_line_number == lineNumber)
                        {
                            headerFileBuilder.Append(f.GetHeaderDefinition());
                            headerFileBuilder.Append(Environment.NewLine);
                            headerFileBuilder.Append(Environment.NewLine);

                            sourceFileStubBuilder.Append(f.GetFunctionStub(""));
                            sourceFileStubBuilder.Append(Environment.NewLine);
                            sourceFileStubBuilder.Append(Environment.NewLine);
                        }
                    }
                }

                if (headerAndSourceAreSameFile)
                {
                    headerFileBuilder.Append("%STUBCONTENT%");
                    headerFileBuilder.Append(Environment.NewLine);
                }
                headerFileBuilder.Append(String.Format("#endif", headerFileDefString));

                string parentPath = Path.GetDirectoryName(stubSourceFilePath);
                if (!Directory.Exists(parentPath))
                {
                    try
                    {
                        Directory.CreateDirectory(parentPath);
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(String.Format("Could not create a required output path ('{0}'): {1}", parentPath, ex.Message));
                        Environment.Exit(5);
                    }
                }
                string headerFileContent = headerFileBuilder.ToString();
                string sourceFileStubContent = sourceFileStubBuilder.ToString();
                if (headerAndSourceAreSameFile)
                {
                    headerFileContent = headerFileContent.Replace("%STUBCONTENT%", sourceFileStubContent);
                    WriteToOutput(headerFilePath, headerFileContent, FileMode.Create);
                }
                else
                {
                    WriteToOutput(headerFilePath, headerFileContent, FileMode.Create);
                    WriteToOutput(stubSourceFilePath, sourceFileStubContent, FileMode.Create);
                }

                manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.DuplicateEnumNames, "enum", "Header", headerFilePath, headerFileContent, "which is defined multiple times");
                manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.DuplicateStructNames, "struct", "Header", headerFilePath, headerFileContent, "which is defined multiple times");
                manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.DuplicateUnionNames, "union", "Header", headerFilePath, headerFileContent, "which is defined multiple times");

                manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.EnumStructDupeNames, "enum or struct", "Header", headerFilePath, headerFileContent, "which is used as a name for both an enum and a struct");
                manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.EnumUnionDupeNames, "enum or union", "Header", headerFilePath, headerFileContent, "which is used as a name for both an enum and a union");
                manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.StructUnionDupeNames, "struct or union", "Header", headerFilePath, headerFileContent, "which is used as a name for both a struct and a union");

                manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.DuplicateEnumNames, "enum", "Source", stubSourceFilePath, sourceFileStubContent, "which is defined multiple times");
                manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.DuplicateStructNames, "struct", "Source", stubSourceFilePath, sourceFileStubContent, "which is defined multiple times");
                manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.DuplicateUnionNames, "union", "Source", stubSourceFilePath, sourceFileStubContent, "which is defined multiple times");

                manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.EnumStructDupeNames, "enum or struct", "Header", stubSourceFilePath, sourceFileStubContent, "which is used as a name for both an enum and a struct");
                manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.EnumUnionDupeNames, "enum or union", "Header", stubSourceFilePath, sourceFileStubContent, "which is used as a name for both an enum and a union");
                manualChangesBuilder = CheckDuplicateNames(manualChangesBuilder, mhgr.StructUnionDupeNames, "struct or union", "Header", stubSourceFilePath, sourceFileStubContent, "which is used as a name for both a struct and a union");
            }

            // create a list of known manual actions that need to be taken by the user
            manualChangesBuilder.Append(Environment.NewLine);
            WriteToOutput(manualChangesPath, manualChangesBuilder.ToString(), FileMode.Create);

            Console.WriteLine("");
        }

        protected static string GetFunctionPointerTypeDefFromParameter(CParameter param)
        {
            //string pTypeName = param.type_name;
            //if (pTypeName.EndsWith(" *"))
            //{
            //    pTypeName = pTypeName.Substring(0, pTypeName.Length - 2);
            //}
            //string result = string.Format("typedef {0} (*{1})();", pTypeName, param.name);
            string result = string.Format("typedef {0};", param.parameter_signature);
            return result;
        }

        protected static string GetFunctionPointerRefInGhidraFormatFromParameter(CParameter param)
        {
            string result = string.Format("{0} *", param.name);
            return result;
        }

        protected static StringBuilder CheckDuplicateNames(StringBuilder builder, List<string> dupeList, string objectType, string fileType, string filePath, string fileContent, string conflictType)
        {
            foreach (string dn in dupeList)
            {
                Regex rex = new Regex(string.Format("{0}\\s+[0-9A-Za-z_]", dn));
                if (rex.IsMatch(fileContent))
                {
                    builder.Append(string.Format("{0} file '{1}' appears to refer to the {2} '{3}', {4}. This may help you manually resolve the conflict.", fileType, filePath, objectType, dn, conflictType));
                    builder.Append(Environment.NewLine);
                }
            }
            return builder;
        }

        protected struct GenerateFunctionSectionsResult
        {
            public StringBuilder FunctionImportBuilder;
            public StringBuilder FunctionSubImportBuilder;
            public int FunctionNumber;
            public int SubRoutineNumber;
            public int SubroutineFunctionCount;
        }

        protected static GenerateFunctionSectionsResult GenerateFunctionSections(CFunction func, StringBuilder functionImportBuilder,
            StringBuilder functionSubImportBuilder, StringBuilder functionSubSubImportBuilder, int functionNum, int subRoutineNumber, int subroutineFunctionCount,
            string ghidraDefineFunctionsSubroutineTemplate, bool replaceFunctionPointersWithVoidPointers, string prefix, bool assignReturnAddressStorage, bool assignParameterStorage, uint snGPValue, bool assumeSNGPValue)
        {
            StringBuilder subBuilder = new StringBuilder();
            //StringBuilder fallbackBuilder = new StringBuilder();
            string functionSuffix = String.Format("{0}Function{1:X8}", prefix, functionNum);
            //bool allParamsHaveExplicitStorage = true;
            //subBuilder.Append("\t\ttry");
            //subBuilder.Append(Environment.NewLine);
            //subBuilder.Append("\t\t{");
            //subBuilder.Append(Environment.NewLine);
            subBuilder.Append(String.Format("\t\t\t// 0x{0:X8}: {1}", func.offset, func.method_signature));
            subBuilder.Append(Environment.NewLine);

            string returnTypeName = func.return_value.type_name;
            //if (func.return_value.is_function_pointer)
            //{
            //    //if (replaceFunctionPointersWithVoidPointers)
            //    //{
            //    //    returnTypeName = "void *";
            //    //}
            //    returnTypeName = GetFunctionPointerRefInGhidraFormatFromParameter(func.return_value);
            //}
            //returnTypeName = func.return_value.type_name;

            string returnValueVarName = string.Format("returnVal{0}", functionSuffix);
            string parameterArrayVarName = string.Format("functionParams{0}", functionSuffix);
            string returnValIsRegisterParam = "false";
            if ((func.return_value.register_name != null) && (func.return_value.register_name != ""))
            {
                returnValIsRegisterParam = "true";
            }

            string returnValueTypeNameVar = string.Format("returnType{0}", functionSuffix);
            subBuilder.Append(String.Format("\t\t\tDataType {0} = GetDataTypeFromMap(dataTypeMap, \"{1}\");", returnValueTypeNameVar, returnTypeName));
            subBuilder.Append(Environment.NewLine);
            subBuilder.Append(String.Format("\t\t\tFunctionParameter {0} = new FunctionParameter(\"{1}\", {2}, \"{3}\", {4}, {5});", returnValueVarName, func.return_value.name, returnValueTypeNameVar, CFunction.GetGhidrafiedRegisterName(func.return_value), func.return_value.stack_offset, returnValIsRegisterParam));
            subBuilder.Append(Environment.NewLine);
            subBuilder.Append(String.Format("\t\t\tFunctionParameter[] {0} = new FunctionParameter[{1}];", parameterArrayVarName, func.parameters.Count));
            subBuilder.Append(Environment.NewLine);
            for (int i = 0; i < func.parameters.Count; i++)
            {
                string isRegisterParam = "false";
                if ((func.parameters[i].register_name != null) && (func.parameters[i].register_name != ""))
                {
                    isRegisterParam = "true";
                }
                string paramTypeName = func.parameters[i].type_name;
                //if (func.parameters[i].is_function_pointer)
                //{
                //    paramTypeName = GetFunctionPointerRefInGhidraFormatFromParameter(func.parameters[i]);
                //}
                string parameterName = func.parameters[i].name;
                string paramTypeVar = string.Format("paramType{0}{1}", functionSuffix, i);
                List<int> arraySizes = new List<int>();
                if (paramTypeName.Contains("["))
                {
                    if (paramTypeName.Contains("])"))
                    {
                        Console.WriteLine(string.Format("Error: the parameter '{0}' in function '{1}' is too complex for this version of TDR to handle automatically. You will need to manually define this function parameter in Ghidra.", func.parameters[i].parameter_signature, func.method_signature));
                    }
                    else
                    {
                        Regex arrayRegex = new Regex("\\[(?<arraySize>[0-9]+)\\]");
                        MatchCollection arrayMatches = arrayRegex.Matches(paramTypeName);
                        foreach (Match m in arrayMatches)
                        {
                            if (m.Groups["arraySize"] != null)
                            {
                                try
                                {
                                    int sz = int.Parse(m.Groups["arraySize"].Value);
                                    arraySizes.Add(sz);
                                }
                                catch (Exception ex)
                                {
                                    Console.WriteLine(string.Format("Error: could not parse array size '{0}' from parameter type '{1}' in function with signature '{2}'", m.Groups["arraySize"].Value, paramTypeName, func.method_signature));
                                }
                            }
                        }
                        paramTypeName = paramTypeName.Substring(0, paramTypeName.IndexOf("["));
                        // otherwise multidimensional arrays will be backwards
                        arraySizes.Reverse();
                    }
                }
                string originalParamTypeName = paramTypeName;
                bool handledParamType = false;
                if (paramTypeName.EndsWith("(*)"))
                {
                    // pointer to something (probably an array)
                    paramTypeName = paramTypeName.Substring(0, paramTypeName.Length - 3).Trim();
                    subBuilder.Append(String.Format("\t\t\tDataType {0} = GetDataTypeFromMap(dataTypeMap, \"{1}\");", paramTypeVar, paramTypeName));
                    subBuilder.Append(Environment.NewLine);
                    GetArrayWrapperResult awr = GetArrayWrapper(subBuilder, arraySizes, paramTypeVar);
                    subBuilder = awr.SubBuilder;
                    paramTypeVar = awr.NewOuterVariableName;
                    string pointerTypeName = string.Format("{0}Pointer", paramTypeVar);
                    subBuilder.Append(String.Format("\t\t\tPointerDataType {0} = new PointerDataType({1});", pointerTypeName, paramTypeVar));
                    subBuilder.Append(Environment.NewLine);
                    paramTypeVar = pointerTypeName;
                    handledParamType = true;
                }
                if ((!handledParamType) && (paramTypeName.EndsWith("*")))
                {
                    // it's either a pointer to a single thing, or an array of pointers to something else
                    if (arraySizes.Count == 0)
                    {
                        // pointer to something
                        subBuilder.Append(String.Format("\t\t\tDataType {0} = GetDataTypeFromMap(dataTypeMap, \"{1}\");", paramTypeVar, paramTypeName));
                        subBuilder.Append(Environment.NewLine);
                    }
                    else
                    {
                        // array of pointers
                        paramTypeName = paramTypeName.Substring(0, paramTypeName.Length - 1).Trim();
                        subBuilder.Append(String.Format("\t\t\tDataType {0} = GetDataTypeFromMap(dataTypeMap, \"{1}\");", paramTypeVar, paramTypeName));
                        subBuilder.Append(Environment.NewLine);
                        string pointerTypeName = string.Format("{0}Pointer", paramTypeVar);
                        subBuilder.Append(String.Format("\t\t\tPointerDataType {0} = new PointerDataType({1});", pointerTypeName, paramTypeVar));
                        subBuilder.Append(Environment.NewLine);
                        GetArrayWrapperResult awr = GetArrayWrapper(subBuilder, arraySizes, paramTypeVar);
                        subBuilder = awr.SubBuilder;
                        paramTypeVar = awr.NewOuterVariableName;
                        paramTypeVar = pointerTypeName;
                    }
                    handledParamType = true;
                }
                if (!handledParamType)
                {
                    subBuilder.Append(String.Format("\t\t\tDataType {0} = GetDataTypeFromMap(dataTypeMap, \"{1}\");", paramTypeVar, paramTypeName));
                    subBuilder.Append(Environment.NewLine);
                }
                subBuilder.Append(String.Format("\t\t\t{0}[{1}] = new FunctionParameter(\"{2}\", {3}, \"{4}\", {5}, {6});", parameterArrayVarName, i, parameterName, paramTypeVar, CFunction.GetGhidrafiedRegisterName(func.parameters[i]), func.parameters[i].stack_offset, isRegisterParam));
                subBuilder.Append(Environment.NewLine);
            }
            string assignRAStorage = "false";
            if (assignReturnAddressStorage)
            {
                assignRAStorage = "true";
            }
            string assignAnyStorage = "false";
            if (assignParameterStorage)
            {
                assignAnyStorage = "true";
            }
            subBuilder.Append(String.Format("\t\t\tDefineFunction2(dataTypeMap, 0x{0:X8}L, \"{1}\", {2}, {3}, {4}, {5});", func.offset, func.name, returnValueVarName, parameterArrayVarName, assignRAStorage, assignAnyStorage));
            subBuilder.Append(Environment.NewLine);
            if (assumeSNGPValue)
            {
                subBuilder.Append(String.Format("\t\t\tSetRegisterAtAddress(ctx, \"gp\", 0x{0:X8}L, 0x{1:X8}L);", func.offset, snGPValue));
                subBuilder.Append(Environment.NewLine);
            }
            subBuilder.Append(Environment.NewLine);

            // only add this block of lines if the entire set of operations for this function succeeded
            functionSubSubImportBuilder.Append(subBuilder.ToString());
            functionNum++;
            subroutineFunctionCount++;
            if (subroutineFunctionCount > MaxFunctionDefinitionsPerGhidraScriptSubroutine)
            {
                string subDefinition = ghidraDefineFunctionsSubroutineTemplate.Replace("%FUNCTION_IMPORT_SUB_CONTENT%", functionSubSubImportBuilder.ToString());
                functionSubSubImportBuilder.Clear();
                subDefinition = subDefinition.Replace("%PREFIX%", prefix);
                subDefinition = subDefinition.Replace("%DEFINE_FUNCTIONS_SUBROUTINE_NUMBER%", String.Format("{0}", subRoutineNumber));
                functionSubImportBuilder.Append(subDefinition);
                functionImportBuilder.Append(String.Format("\t\t{0}DefineFunctionsSub{1}(dataTypeMap, ctx);", prefix, subRoutineNumber));
                functionImportBuilder.Append(Environment.NewLine);
                subRoutineNumber++;
                subroutineFunctionCount = 0;
            }

            GenerateFunctionSectionsResult result = new GenerateFunctionSectionsResult();
            result.FunctionImportBuilder = functionImportBuilder;
            result.FunctionSubImportBuilder = functionSubImportBuilder;
            result.FunctionNumber = functionNum;
            result.SubRoutineNumber = subRoutineNumber;
            result.SubroutineFunctionCount = subroutineFunctionCount;

            return result;
        }

        protected static GenerateFunctionSectionsResult FlushLastEntriesForFunctionSections(StringBuilder functionImportBuilder,
            StringBuilder functionSubImportBuilder, StringBuilder functionSubSubImportBuilder, int functionNum, int subRoutineNumber, int subroutineFunctionCount,
            string ghidraDefineFunctionsSubroutineTemplate, string prefix)
        {
            StringBuilder subBuilder = new StringBuilder();
            string functionSuffix = String.Format("{0}Function{1:X8}", prefix, functionNum);

            string subDefinition = ghidraDefineFunctionsSubroutineTemplate.Replace("%FUNCTION_IMPORT_SUB_CONTENT%", functionSubSubImportBuilder.ToString());
            functionSubSubImportBuilder.Clear();
            subDefinition = subDefinition.Replace("%PREFIX%", prefix);
            subDefinition = subDefinition.Replace("%DEFINE_FUNCTIONS_SUBROUTINE_NUMBER%", String.Format("{0}", subRoutineNumber));
            functionSubImportBuilder.Append(subDefinition);
            functionImportBuilder.Append(String.Format("\t\t{0}DefineFunctionsSub{1}(dataTypeMap, ctx);", prefix, subRoutineNumber));
            functionImportBuilder.Append(Environment.NewLine);
            subRoutineNumber++;
            subroutineFunctionCount = 0;

            GenerateFunctionSectionsResult result = new GenerateFunctionSectionsResult();
            result.FunctionImportBuilder = functionImportBuilder;
            result.FunctionSubImportBuilder = functionSubImportBuilder;
            result.FunctionNumber = functionNum;
            result.SubRoutineNumber = subRoutineNumber;
            result.SubroutineFunctionCount = subroutineFunctionCount;

            return result;
        }

        public struct GetArrayWrapperResult
        {
            public StringBuilder SubBuilder;
            public string NewOuterVariableName;
        }

        protected static GetArrayWrapperResult GetArrayWrapper(StringBuilder subBuilder, List<int> arraySizes, string paramTypeVar)
        {
            if (arraySizes.Count > 0)
            {
                string baseTypeName = paramTypeVar;
                for (int arrayNum = 0; arrayNum < arraySizes.Count; arrayNum++)
                {
                    string currentArrayTypeName = paramTypeVar;
                    if (arrayNum > 0)
                    {
                        currentArrayTypeName = string.Format("{0}Array{1}", baseTypeName, arrayNum - 1);
                    }
                    string newArrayTypeName = string.Format("{0}Array{1}", baseTypeName, arrayNum);
                    subBuilder.Append(String.Format("\t\t\tArrayDataType {0} = new ArrayDataType({1}, {2}, {3}.getLength());", newArrayTypeName, currentArrayTypeName, arraySizes[arrayNum], baseTypeName));
                    subBuilder.Append(Environment.NewLine);
                    paramTypeVar = newArrayTypeName;
                }
            }

            GetArrayWrapperResult result = new GetArrayWrapperResult();
            result.SubBuilder = subBuilder;
            result.NewOuterVariableName = paramTypeVar;
            return result;
        }

        static void DisplayHelpMessage()
        {
            DisplayBanner();

            Console.Write(GetEmbeddedTextFile("CreateSkeleton.HelpMessage.txt"));
            Console.WriteLine();
        }

        protected static string GetEmbeddedTextFile(string fileName)
        {
            Assembly asm = Assembly.GetExecutingAssembly();
            Stream inStream = asm.GetManifestResourceStream(fileName);
            StreamReader inReader = new StreamReader(inStream);
            return inReader.ReadToEnd();
        }

        static void DisplayBanner()
        {
            Assembly currentAssembly = Assembly.GetExecutingAssembly();

            AssemblyName assemblyName = currentAssembly.GetName();

            int versionMajor = assemblyName.Version.Major;
            int versionMinor = assemblyName.Version.Minor;

            Console.WriteLine(string.Format("{0} {1}.{2}", ProgramName, versionMajor, versionMinor));
            Console.WriteLine(Copyright);
            Console.WriteLine(InfoURL);
            Console.WriteLine(Environment.NewLine);
        }

        protected static void ExitWithError(string message, int exitCode)
        {
            DisplayBanner();
            Console.WriteLine(message);
            System.Environment.ExitCode = exitCode;
            Environment.Exit(exitCode);
        }

        private static void WriteToOutput(string outputPath, string text, FileMode mode)
        {
            FileStream oStream;
            StreamWriter oWriter;

            oStream = new FileStream(outputPath, mode, FileAccess.Write);
            oWriter = new StreamWriter(oStream);

            oWriter.WriteLine(text);

            oWriter.Close();
            oStream.Close();
        }

        private static string GetFileAsString(string inputPath)
        {
            string result = "";
            FileStream inStream;
            StreamReader inReader;

            inStream = new FileStream(inputPath, FileMode.Open, FileAccess.Read);
            inReader = new StreamReader(inStream);

            result = inReader.ReadToEnd();

            inReader.Close();
            inStream.Close();
            return result;
        }

        private static string[] GetLinesAsStringArray(string inputString)
        {
            // for easier cross-platform use, this handles Unix and DOS-style newlines the same way
            string inString = inputString.Replace("\r", "");

            string[] result = inString.Split(new string[] { "\n" }, StringSplitOptions.RemoveEmptyEntries);

            return result;
        }

        private static string[] GetFileLinesAsStringArray(string inputPath)
        {
            return GetLinesAsStringArray(GetFileAsString(inputPath));
        }
    }
}
