/* ###
 * IP: GHIDRA
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
// based in part on the IterateDataScript included with Ghidra

import java.io.*;
import java.util.*;

import ghidra.app.plugin.core.script.Ingredient;
import ghidra.app.plugin.core.script.IngredientDescription;
import ghidra.app.script.GatherParamPanel;
import ghidra.app.script.GhidraScript;
import ghidra.program.model.data.*;
import ghidra.program.model.listing.Data;
import ghidra.program.model.symbol.SymbolUtilities;

import utilities.util.FileUtilities;

public class TDRIterateDataScript extends GhidraScript implements Ingredient {

	int _MaxRecursionDepth = 20;

	@Override
	public void run() throws Exception {
		IngredientDescription[] ingredients = getIngredientDescriptions();
		for (IngredientDescription ingredient : ingredients) {
			state.addParameter(ingredient.getID(), ingredient.getLabel(), ingredient.getType(),
				ingredient.getDefaultValue());
		}
		if (!state.displayParameterGatherer("Script Options")) {
			return;
		}
		File outputFile = (File) state.getEnvironmentVar("EmbeddedDataOutputFile");
		FileWriter output = new FileWriter(outputFile);
		BufferedWriter writer = new BufferedWriter(output);

		Data data = getFirstData();

		while (true) {

			if (monitor.isCancelled()) {
				break;
			}

			if (data == null) {
				break;
			}

			writer.write(getDataDescription(data, 0, _MaxRecursionDepth));
			writer.newLine();
			data = getDataAfter(data);
		}
		
		writer.close();
	}
	
	protected String getDataDescription(Data data, int recursionDepth, int maxRecursionDepth)
	{
		if (recursionDepth >= maxRecursionDepth)
		{
			println("Error: hit recursion depth limit of " + String.valueOf(recursionDepth) + " while traversing this item");
			return "";
		}
		DataType dt = data.getDataType();
		DataType[] dtParents = dt.getParents();
		/*
		StringBuffer parentBuffer = new StringBuffer();
		parentBuffer.append("[");
		for (int parentNum = 0; parentNum < dtParents.length; parentNum++)
		{
			parentBuffer.append(dtParents[parentNum].getName());
			if (parentNum < (dtParents.length - 1))
			{
				parentBuffer.append(", ");
			}
		}
		parentBuffer.append("]");
		*/

		StringBuffer buffer = new StringBuffer();

		if (recursionDepth > 0)
		{
			for (int i = 0; i < recursionDepth; i++)
			{
				buffer.append("\t");
			}
		}

		buffer.append("0x");
		buffer.append(data.getMinAddress());
		buffer.append("\t");
		buffer.append(String.format("0x%x", dt.getLength()));
		buffer.append("\t");
		buffer.append(SymbolUtilities.getDynamicName​(currentProgram, data.getMinAddress()));
		buffer.append("\t");
		buffer.append(data.getFieldName());
		buffer.append("\t");
		buffer.append(dt.getName());
		buffer.append("\t");
		buffer.append(dt.getDisplayName());
		buffer.append("\t");
		//buffer.append(parentBuffer);
		//buffer.append("\t");
		buffer.append(data.getMnemonicString());
		buffer.append("\t");
		buffer.append(data.getValue());
		int subComponentCount = data.getNumComponents();
		if (subComponentCount > 0)
		{
			for (int scNum = 0; scNum < subComponentCount; scNum++)
			{
				buffer.append(System.lineSeparator());
				buffer.append(getDataDescription(data.getComponent(scNum), recursionDepth + 1, maxRecursionDepth));
			}
		}

		return buffer.toString();
	}

	@Override
	public IngredientDescription[] getIngredientDescriptions() {
		IngredientDescription[] retVal = new IngredientDescription[] {
			new IngredientDescription("EmbeddedDataOutputFile", "Output File", GatherParamPanel.FILE, "F:\\Decompile\\Data_Export_Test.txt") };
		return retVal;
	}
	
}
