﻿// this entire project/solution is a heavily-customized fork of https://github.com/stohrendorf/symdump/

// SymDumpTE
// Exports debug symbol data from PsyQ .SYM files.
// Original code by Steffen Ohrendorf (https://github.com/stohrendorf/symdump/)
// Customizations by Ben Lincoln, 2019
// https://www.beneaththewaves.net/

// This file is part of SymDumpTE.

// SymDumpTE is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// SymDumpTE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with SymDumpTE (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

using symdump.exefile;
using symdump.exefile.util;
using symdump.symfile;

namespace SymDumpTE
{
    class Program
    {
        protected const string ProgramName = "SymDumpTE";
        protected static string Copyright = "Original SymDump by Steffen Ohrendorf, extended and customized into SymDumpTE by Ben Lincoln, 2019";
        protected const string InfoURL = "Original: https://github.com/stohrendorf/symdump/, Customized: https://www.beneaththewaves.net/Software/";

        static void Main(string[] args)
        {
            if ((args.Length < 1))
            {
                DisplayHelpMessage();
                Environment.Exit(1);
            }

            string inputFilePath = "";
            string outputFilePath = "";
            bool jsonOutput = false;
            bool disassemble = false;
            bool autoRenameFakes = false;
            bool renameForCompatibility = false;

            ArrayList arguments = new ArrayList();
            for (int i = 0; i < args.Length; i++)
            {
                // if the argument is one with a second component, catch that now
                //string nextArg = "";
                switch (args[i])
                {
                    default:
                        arguments.Add(args[i]);
                        break;
                }
            }

            ArrayList positionalArgs = new ArrayList();

            foreach (string arg in arguments)
            {
                if ((arg == "-?") || (arg == "-h") || (arg == "--help"))
                {
                    DisplayHelpMessage();
                    Environment.Exit(0);
                }
                bool handled = false;
                if (arg == "--json")
                {
                    jsonOutput = true;
                    handled = true;
                }
                if (arg == "--disassemble")
                {
                    disassemble = true;
                    handled = true;
                }
                if (arg == "--auto-rename-fakes")
                {
                    autoRenameFakes = true;
                    handled = true;
                }
                if (arg == "--rename-for-compatibility")
                {
                    renameForCompatibility = true;
                    handled = true;
                }
                if (!handled)
                {
                    positionalArgs.Add(arg);
                }
            }

            if (positionalArgs.Count > 0)
            {
                inputFilePath = (string)positionalArgs[0];
            }

            if (positionalArgs.Count > 1)
            {
                outputFilePath = (string)positionalArgs[1];
            }

            if (inputFilePath == "")
            {
                ExitWithError("No input symbol file path was specified.", 1);
            }

            if (!File.Exists(inputFilePath))
            {
                ExitWithError("The specified input symbol file does not exist.", 1);
            }

            SymFile symFile;

            var exeFilename = Path.ChangeExtension(inputFilePath, "EXE");

            using (var fs = new FileStream(inputFilePath, FileMode.Open))
            {
                symFile = new SymFile(new BinaryReader(fs), renameForCompatibility);
                string defaultSourceFileBaseName = Path.GetFileNameWithoutExtension(inputFilePath);

                if (File.Exists(exeFilename))
                {
                    try
                    {
                        symFile.GetDataFromPSXEXE(exeFilename);
                    }
                    catch (Exception ex)
                    {
                        ExitWithError(String.Format("Could not get data from PSX EXE '{0}': {1}", exeFilename, ex.Message), 9);
                    }
                }

                if (jsonOutput)
                {
                    string result = symFile.DumpJSON(defaultSourceFileBaseName, autoRenameFakes);
                    if (outputFilePath == "")
                    {
                        Console.WriteLine(result);
                    }
                    else
                    {
                        try
                        {
                            WriteToOutput(outputFilePath, result, FileMode.Create);
                        }
                        catch (Exception ex)
                        {
                            ExitWithError(String.Format("Could not write to the output file '{0}': {1}", outputFilePath, ex.Message), 7);
                        }
                    }
                }
                else
                {
                    symFile.Dump(System.Console.Out);
                }
            }

            if (disassemble)
            {
                if (!File.Exists(exeFilename))
                {
                    System.Console.WriteLine(String.Format("// could not find file '{0}' to disassemble", exeFilename));
                    return;
                }

                System.Console.WriteLine();
                System.Console.WriteLine("// disassembly");

                using (var fs = new EndianBinaryReader(new FileStream(exeFilename, FileMode.Open)))
                {
                    var exeFile = new ExeFile(fs, symFile);
                    exeFile.Disassemble();
                    exeFile.Dump();
                }
            }
            Console.WriteLine();
        }

        static void DisplayHelpMessage()
        {
            DisplayBanner();

            Assembly asm = Assembly.GetExecutingAssembly();
            Stream inStream = asm.GetManifestResourceStream("SymDumpTE.HelpMessage.txt");
            StreamReader inReader = new StreamReader(inStream);
            Console.Write(inReader.ReadToEnd());
            Console.WriteLine();
            inReader.Close();
            inStream.Close();
        }

        static void DisplayBanner()
        {
            Assembly currentAssembly = Assembly.GetExecutingAssembly();

            AssemblyName assemblyName = currentAssembly.GetName();

            int versionMajor = assemblyName.Version.Major;
            int versionMinor = assemblyName.Version.Minor;

            Console.WriteLine(string.Format("{0} {1}.{2}", ProgramName, versionMajor, versionMinor));
            Console.WriteLine(Copyright);
            Console.WriteLine(InfoURL);
            Console.WriteLine(Environment.NewLine);
        }

        protected static void ExitWithError(string message, int exitCode)
        {
            DisplayBanner();
            Console.WriteLine(message);
            System.Environment.ExitCode = exitCode;
            Environment.Exit(exitCode);
        }

        private static void WriteToOutput(string outputPath, string text, FileMode mode)
        {
            FileStream oStream;
            StreamWriter oWriter;

            oStream = new FileStream(outputPath, mode, FileAccess.Write);
            oWriter = new StreamWriter(oStream);

            oWriter.WriteLine(text);

            oWriter.Close();
            oStream.Close();
        }
    }
}
