﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using symdump.util;

namespace symdump.symfile
{
    public class StructDef : IEquatable<StructDef>
    {
        private List<StructMember> _members = new List<StructMember>();
        private string _name;
        public readonly int SYMFileSequenceNumber;

        public string Name
        {
            get
            {
                return _name;
            }
            set
            {
                _name = value;
            }
        }

        public List<StructMember> Members
        {
            get
            {
                return _members;
            }
            set
            {
                _members = value;
            }
        }

        public StructDef(BinaryReader stream, string name, BenLincoln.DebugSymbols.SequenceIssuer issuer, bool renameForCompatibility)
        {
            _name = name;
            if (renameForCompatibility)
            {
                _name = BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(_name);
            }
            SYMFileSequenceNumber = issuer.GetNextNumber();
            while (true)
            {
                var typedValue = new TypedValue(stream, issuer);
                if (typedValue.Type == (0x80 | 20))
                {
                    var m = new StructMember(typedValue, stream, false, string.Format("struct {0}", this.Name), this.GetHashCode(), issuer, renameForCompatibility);

                    if (m.TypeInfo.ClassType == ClassType.EndOfStruct)
                        break;

                    _members.Add(m);
                }
                else if (typedValue.Type == (0x80 | 22))
                {
                    var m = new StructMember(typedValue, stream, true, string.Format("struct {0}", this.Name), this.GetHashCode(), issuer, renameForCompatibility);

                    if (m.TypeInfo.ClassType == ClassType.EndOfStruct)
                        break;

                    _members.Add(m);
                }
                else
                {
                    Console.WriteLine(string.Format("Unexpected struct type: {0} - data will be ignored", typedValue.Type));
                }
            }
        }

        public bool IsFake => new Regex(@"^(\.|_)\d+fake$").IsMatch(_name);

        public bool Equals(StructDef other)
        {
            if (ReferenceEquals(null, other)) return false;
            if (ReferenceEquals(this, other)) return true;
            return _members.SequenceEqual(other._members) && string.Equals(_name, other._name);
        }

        public void Dump(IndentedTextWriter writer)
        {
            writer.WriteLine($"struct {_name} {{");
            ++writer.Indent;
            foreach (var m in _members)
                writer.WriteLine(m);
            --writer.Indent;
            writer.WriteLine("};");
        }

        public BenLincoln.DebugSymbols.SymDump.CStruct ToBLDSCStruct()
        {
            BenLincoln.DebugSymbols.SymDump.CStruct result = new BenLincoln.DebugSymbols.SymDump.CStruct();
            result.name = _name;
            result.hashcode = GetHashCode();
            result.sym_file_sequence_number = SYMFileSequenceNumber;
            result.members = new List<BenLincoln.DebugSymbols.SymDump.CStructMember>();
            foreach (StructMember m in _members)
            {
                result.members.Add(m.ToBLDSCStructMember());
            }

            return result;
        }

        public string DumpJSON(string sourceFile, string lineIndent)
        {
            BenLincoln.DebugSymbols.SymDump.CStruct resultDef = ToBLDSCStruct();

            return BenLincoln.DebugSymbols.Utilities.IndentString(resultDef.ToJSON(), lineIndent);
        }

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj)) return false;
            if (ReferenceEquals(this, obj)) return true;
            if (obj.GetType() != GetType()) return false;
            return Equals((StructDef) obj);
        }

        public override int GetHashCode()
        {
            unchecked
            {
                return ((_members != null ? _members.GetHashCode() : 0) * 397) ^
                       (_name != null ? _name.GetHashCode() : 0);
            }
        }

        public StructMember ForOffset(uint ofs)
        {
            return _members
                .Where(m => m.TypeInfo.ClassType != ClassType.Bitfield && m.TypedValue.Value <= ofs)
                .OrderBy(m => m.TypedValue.Value)
                .FirstOrDefault();
        }
    }
}
