﻿using System;
using System.IO;
using symdump.symfile.util;

namespace symdump.symfile
{
    public class StructMember : IEquatable<StructMember>
    {
        private readonly string _name;
        public readonly TypedValue TypedValue;
        public readonly TypeInfo TypeInfo;
        public readonly string ParentName;
        public readonly int ParentHashCode;

        public StructMember(TypedValue tv, BinaryReader reader, bool extended, string parentName, int parentHashCode, BenLincoln.DebugSymbols.SequenceIssuer issuer, bool renameForCompatibility)
        {
            TypeInfo = reader.ReadTypeInfo(extended, issuer, renameForCompatibility);
            _name = reader.ReadPascalString();
            if (renameForCompatibility)
            {
                _name = BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(_name);
            }
            TypedValue = tv;
            ParentName = parentName;
            ParentHashCode = parentHashCode;

            if (TypeInfo.ClassType != ClassType.Bitfield && TypeInfo.ClassType != ClassType.StructMember &&
                TypeInfo.ClassType != ClassType.UnionMember && TypeInfo.ClassType != ClassType.EndOfStruct)
                throw new Exception($"Unexpected class {TypeInfo.ClassType}");
        }

        public bool Equals(StructMember other)
        {
            if (ReferenceEquals(null, other)) return false;
            if (ReferenceEquals(this, other)) return true;
            return string.Equals(_name, other._name) && TypeInfo.Equals(other.TypeInfo) &&
                   TypedValue.Equals(other.TypedValue);
        }

        public override string ToString()
        {
            switch (TypeInfo.ClassType)
            {
                case ClassType.Bitfield:
                    return TypeInfo.AsCode(_name) +
                           $" : {TypeInfo.Size}; // offset={TypedValue.Value / 8}.{TypedValue.Value % 8}";
                case ClassType.StructMember:
                case ClassType.UnionMember:
                    return TypeInfo.AsCode(_name) +
                           $"; // size={TypeInfo.Size}, offset={TypedValue.Value}";
                default:
                    throw new Exception($"Unexpected class {TypeInfo.ClassType}");
            }
        }

        public string GetClassTypeName()
        {
            switch (TypeInfo.ClassType)
            {
                case ClassType.Bitfield:
                    return "bitfield";
                case ClassType.StructMember:
                    return "struct_member";
                case ClassType.UnionMember:
                    return "union_member";
                default:
                    break;
            }
            throw new Exception($"Unexpected class {TypeInfo.ClassType}");
        }

        public BenLincoln.DebugSymbols.SymDump.CStructMember ToBLDSCStructMember()
        {
            BenLincoln.DebugSymbols.SymDump.CStructMember result = new BenLincoln.DebugSymbols.SymDump.CStructMember();
            result.name = _name;
            //result.c_type = TypeInfo.GetCType();
            //result.type_name = TypeInfo.GetName(_name);
            result.struct_member_signature = ToString();
            //BenLincoln.DebugSymbols.SymDump.CParameter[] pParsed = BenLincoln.DebugSymbols.SymDump.CElement.GetCParametersFromString(result.struct_member_signature, true);
            //result.c_type = pParsed[0].type_name;
            //result.type_name = pParsed[0].type_name;
            result.c_type = TypeInfo.AsCode("");
            result.type_name = result.c_type;
            result.parent_name = ParentName;
            //result.hashcode = GetHashCode();
            result.class_type = GetClassTypeName();
            result.size = TypeInfo.Size;
            result.derived_types = TypeInfo.GetDerivedTypes(_name).ToArray();
            result.sym_file_sequence_number = TypedValue.SYMFileSequenceNumber;
            // damn it, Psygnosis, or whoever decided to use 0 for "default" instead of including the actual size
            if (result.size == 0)
            {
                bool handled = false;
                switch(result.c_type)
                {
                    case "byte":
                    case "char":
                    case "unsigned char":
                        result.size = 1;
                        handled = true;
                        break;
                    case "short":
                    case "unsigned short":
                        result.size = 2;
                        handled = true;
                        break;
                    case "float":
                    case "int":
                    case "long":
                    case "unsigned int":
                        result.size = 4;
                        handled = true;
                        break;
                    case "double":
                    case "unsigned long":
                        result.size = 8;
                        handled = true;
                        break;
                    case "void":
                    case "NULL":
                        // not sure this is correct
                        result.size = 4;
                        handled = true;
                        break;
                }
                // catch pointers
                if (!handled)
                {
                    if (result.c_type.EndsWith("*"))
                    {
                        result.size = 4;
                        handled = true;
                    }
                }

                // catch function pointers
                if (!handled)
                {
                    if (result.c_type.EndsWith("(*)()"))
                    {
                        result.size = 4;
                        handled = true;
                    }
                    if (result.c_type.EndsWith("(**)()"))
                    {
                        result.size = 4;
                        handled = true;
                    }
                    if (result.c_type.EndsWith("(***)()"))
                    {
                        result.size = 4;
                        handled = true;
                    }
                }

                // catch zero-size (variable-length) arrays, which, who knows? I'm just going to leave them set to zero, because, whatever, C standards committee
                // and Psygnosis, and anyone else involved in this. Thanks for making things difficult.
                if (!handled)
                {
                    if ((result.c_type.EndsWith("[0]")) || (result.c_type.EndsWith("[0])")))
                    {
                        result.size = 0;
                        handled = true;
                    }
                }

                if (!handled)
                {
                    throw new Exception(String.Format("Unknown type: '{0}' - cannot set size", result.c_type));
                }
            }
            result.offset = TypedValue.Value;

            return result;
        }

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj)) return false;
            if (ReferenceEquals(this, obj)) return true;
            if (obj.GetType() != GetType()) return false;
            return Equals((StructMember) obj);
        }

        public override int GetHashCode()
        {
            unchecked
            {
                var hashCode = _name != null ? _name.GetHashCode() : 0;
                hashCode = (hashCode * 397) ^ (TypeInfo != null ? TypeInfo.GetHashCode() : 0);
                hashCode = (hashCode * 397) ^ (TypedValue != null ? TypedValue.GetHashCode() : 0);
                return hashCode;
            }
        }
    }
}
