using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using symdump.symfile.util;

namespace symdump.symfile
{
    public class TypeInfo : IEquatable<TypeInfo>
    {
        public readonly ClassType ClassType;
        public readonly uint[] Dims;
        public readonly uint Size;
        public readonly string Tag;
        public readonly TypeDef TypeDef;
        public readonly int SYMFileSequenceNumber;

        public TypeInfo(BinaryReader reader, bool extended, BenLincoln.DebugSymbols.SequenceIssuer issuer, bool renameForCompatibility)
        {
            ClassType = reader.ReadClassType();
            TypeDef = reader.ReadTypeDef(issuer);
            Size = reader.ReadUInt32();
            SYMFileSequenceNumber = issuer.GetNextNumber();

            if (extended)
            {
                var n = reader.ReadUInt16();
                Dims = new uint[n];
                for (var i = 0; i < n; ++i)
                    Dims[i] = reader.ReadUInt32();

                Tag = reader.ReadPascalString();
                if (renameForCompatibility)
                {
                    if (Tag.Contains("fake"))
                    {
                        Tag = BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(Tag);
                    }
                }
            }
            else
            {
                Dims = new uint[0];
                Tag = null;
            }
        }

        private bool IsFake => Tag != null && new Regex(@"^(\.|_)\d+fake$").IsMatch(Tag);

        public bool Equals(TypeInfo other)
        {
            if (ReferenceEquals(null, other)) return false;
            if (ReferenceEquals(this, other)) return true;

            if (IsFake)
                return ClassType == other.ClassType && TypeDef.Equals(other.TypeDef) && Size == other.Size &&
                       Dims.SequenceEqual(other.Dims) && other.IsFake;
            return ClassType == other.ClassType && TypeDef.Equals(other.TypeDef) && Size == other.Size &&
                   Dims.SequenceEqual(other.Dims) && string.Equals(Tag, other.Tag);
        }

        public override string ToString()
        {
            return $"classType={ClassType} typeDef={TypeDef} size={Size}, dims=[{string.Join(",", Dims)}]";
        }

        public string AsCode(string name)
        {
            return TypeDef.AsCode(name, this).Trim();
        }

        public string GetCType()
        {
            return TypeDef.GetCType(this, ToString());
        }

        public string GetName(string name)
        {
            return TypeDef.GetName(name, this);
        }

        public List<string> GetDerivedTypes(string name)
        {
            return TypeDef.GetDerivedTypeList(name, this);
        }

        public BenLincoln.DebugSymbols.SymDump.CTypeDef ToBLDSCTypeDef(string name, string sourceFile)
        {
            BenLincoln.DebugSymbols.SymDump.CTypeDef result = new BenLincoln.DebugSymbols.SymDump.CTypeDef();

            result.name = GetName(name);
            result.c_type = GetCType();
            result.source_file = sourceFile;
            //result.hashcode = GetHashCode();
            result.typedef_signature = AsCode(name);

            return result;
        }

        public string DumpJSON(string name, string sourceFile, string lineIndent)
        {
            BenLincoln.DebugSymbols.SymDump.CTypeDef resultDef = ToBLDSCTypeDef(name, sourceFile);

            return BenLincoln.DebugSymbols.Utilities.IndentString(resultDef.ToJSON(), lineIndent);
            //StringBuilder builder = new StringBuilder();

            //builder.Append(lineIndent);
            //builder.Append("{");
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append(String.Format("\t\"name\": \"{0}\",", name));
            //builder.Append(Environment.NewLine);
            //if (sourceFile != "")
            //{
            //    builder.Append(lineIndent);
            //    builder.Append(String.Format("\t\"source_file\": \"{0}\",", sourceFile));
            //    builder.Append(Environment.NewLine);
            //}
            //builder.Append(lineIndent);
            //builder.Append(String.Format("\t\"definition\": \"{0}\",", AsCode(name)));
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append("}");

            //return builder.ToString();
        }

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj)) return false;
            if (ReferenceEquals(this, obj)) return true;
            if (obj.GetType() != GetType()) return false;
            return Equals((TypeInfo) obj);
        }

        public override int GetHashCode()
        {
            unchecked
            {
                var hashCode = (int) ClassType;
                hashCode = (hashCode * 397) ^ TypeDef.GetHashCode();
                hashCode = (hashCode * 397) ^ (int) Size;
                hashCode = (hashCode * 397) ^ Dims.GetHashCode();
                hashCode = (hashCode * 397) ^ (!IsFake ? Tag.GetHashCode() : 0);
                return hashCode;
            }
        }
    }
}
