﻿// PopulateSkeleton
// Populates a skeleton PlayStation dev project based on the output of SymDumpTE and CreateSkeleton
// Copyright 2019 Ben Lincoln
// https://www.beneaththewaves.net/

// This file is part of PopulateSkeleton.

// PopulateSkeleton is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// PopulateSkeleton is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with PopulateSkeleton (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using Newtonsoft.Json;
using BenLincoln.DebugSymbols;
using BenLincoln.DebugSymbols.SymDump;

namespace PopulateSkeleton
{
    public struct DecompiledSourceCodeMatch
    {
        public CFunction DecompiledFunction;
        public float MatchPercent;
    }

    class Program
    {
        protected const string ProgramName = "PopulateSkeleton";
        protected static string Copyright = "Copyright 2019 Ben Lincoln";
        protected const string InfoURL = "https://www.beneaththewaves.net/Software/";
        protected const string ProjectNameValidationRegexString = @"^[A-Za-z0-9_]{1,16}$";
        protected const string ProjectNameExplanation = "alphanumeric characters and underscore (_)";

        static void Main(string[] args)
        {
            if ((args.Length < 5))
            {
                DisplayHelpMessage();
                Environment.Exit(1);
            }

            string projectName = "";
            string baseOutputDirectoryPath = "";
            string inputJSONFilePath = "";
            string inputSourceCodeFilePath = "";
            bool forceOverwrite = false;

            ArrayList arguments = new ArrayList();
            for (int i = 0; i < args.Length; i++)
            {
                // if the argument is one with a second component, catch that now
                string nextArg = "";
                switch (args[i])
                {
                    case "--name":
                        nextArg = args[i + 1];
                        projectName = nextArg;
                        Regex pnRex = new Regex(ProjectNameValidationRegexString);
                        if (!pnRex.IsMatch(projectName))
                        {
                            Console.WriteLine(String.Format("The specified project name is invalid. The name must consist only of {0}.", ProjectNameExplanation));
                            Environment.Exit(1);
                        }
                        i++;
                        break;
                    case "--input-json":
                        nextArg = args[i + 1];
                        inputJSONFilePath = nextArg;
                        i++;
                        break;
                    case "--input-source":
                        nextArg = args[i + 1];
                        inputSourceCodeFilePath = nextArg;
                        i++;
                        break;
                    case "--output":
                        nextArg = args[i + 1];
                        baseOutputDirectoryPath = nextArg;
                        i++;
                        break;
                    default:
                        arguments.Add(args[i]);
                        break;
                }
            }

            ArrayList positionalArgs = new ArrayList();

            foreach (string arg in arguments)
            {
                if ((arg == "-?") || (arg == "-h") || (arg == "--help"))
                {
                    DisplayHelpMessage();
                    Environment.Exit(0);
                }
                bool handled = false;
                if (arg == "--overwrite")
                {
                    forceOverwrite = true;
                    handled = true;
                }
                // right now, all positional arguments are file paths
                if (!handled)
                {
                    positionalArgs.Add(arg);
                }
            }

            if (inputJSONFilePath == "")
            {
                ExitWithError("No input SymDump file path was specified.", 1);
            }

            if (!File.Exists(inputJSONFilePath))
            {
                ExitWithError("The specified input SymDump file does not exist.", 1);
            }

            if (inputSourceCodeFilePath == "")
            {
                ExitWithError("No input decompiled source code file path was specified.", 1);
            }

            if (!File.Exists(inputSourceCodeFilePath))
            {
                ExitWithError("The specified input decompiled source code file does not exist.", 1);
            }

            if (baseOutputDirectoryPath == "")
            {
                ExitWithError("No output directory was specified.", 1);
            }
            if (forceOverwrite)
            {
                if (Directory.Exists(baseOutputDirectoryPath))
                {
                    try
                    {
                        Directory.Delete(baseOutputDirectoryPath, true);
                    }
                    catch (Exception ex)
                    {
                        ExitWithError(String.Format("The output directory '{0}' already exists, and could not be deleted: {1}", baseOutputDirectoryPath, ex.Message), 2);
                    }
                }
            }
            /*
            else
            {
                if (Directory.Exists(baseOutputDirectoryPath))
                {
                    ExitWithError("The specified output directory already exists. Use the --overwrite option to replace it.", 1);
                }
            }
            */

            // create base output directory structure
            string[] outputDirs = new string[] { "source-decompiled" };
            foreach (string od in outputDirs)
            {
                string fullPath = Path.Combine(baseOutputDirectoryPath, od);
                try
                {
                    Directory.CreateDirectory(fullPath);
                }
                catch (Exception ex)
                {
                    ExitWithError(String.Format("Could not create the output directory '{0}': {1}", fullPath, ex.Message), 2);
                }
            }

            string manualChangesPath = Path.Combine(baseOutputDirectoryPath, String.Format("README-{0}-PopulateSkeleton-Manual_Changes_Required.txt", projectName));

            // get SymDumpTE output
            string symDumpRaw = "";
            try
            {
                symDumpRaw = GetFileAsString(inputJSONFilePath);
            }
            catch (Exception ex)
            {
                ExitWithError(String.Format("Could not read the input file '{0}': {1}", inputJSONFilePath, ex.Message), 2);
            }

            // get decompiled source
            string decompiledSource = "";
            try
            {
                decompiledSource = GetFileAsString(inputSourceCodeFilePath);
            }
            catch (Exception ex)
            {
                ExitWithError(String.Format("Could not read the input file '{0}': {1}", inputSourceCodeFilePath, ex.Message), 2);
            }

            StringBuilder manualChangesBuilder = new StringBuilder();

            // fix names that don't work in Ghidra
            Regex dotRenameRex = new Regex("(?<itemName>(union|struct) \\.[^ ;\"$]+)[ ;\"$]");
            MatchCollection drMatches = dotRenameRex.Matches(symDumpRaw);
            Dictionary<string, int> dotNames = new Dictionary<string, int>();
            foreach (Match m in drMatches)
            {
                if (m.Groups["itemName"] != null)
                {
                    string matchText = m.Groups["itemName"].Value;
                    if (!dotNames.ContainsKey(matchText))
                    {
                        dotNames.Add(matchText, 1);
                    }
                    else
                    {
                        dotNames[matchText]++;
                    }
                }
            }

            if (dotNames.Count > 0)
            {
                manualChangesBuilder.Append("The input JSON file contains at least one C element with a name that starts with a '.'. This naming convention is incompatible with Ghidra, so it will be replaced by a leading underscore. You should consider renaming these elements in the source JSON file, and will need to do so if any of them have name conflicts listed later in this file.");
                manualChangesBuilder.Append(Environment.NewLine);
                List<string> keyNames = dotNames.Keys.ToList();
                keyNames.Sort();
                manualChangesBuilder.Append("Elements which will be renamed:");
                manualChangesBuilder.Append(Environment.NewLine);
                foreach (string dotName in keyNames)
                {
                    if (dotNames[dotName] > 1)
                    {
                        manualChangesBuilder.Append(string.Format("\t{0} ({1} occurrences)", dotName, dotNames[dotName]));
                    }
                    else
                    {
                        manualChangesBuilder.Append(string.Format("\t{0}", dotName));
                    }
                    manualChangesBuilder.Append(Environment.NewLine);
                }
                manualChangesBuilder.Append(Environment.NewLine);
            }

            symDumpRaw = symDumpRaw.Replace("union .", "union _");
            symDumpRaw = symDumpRaw.Replace("struct .", "struct _");
            symDumpRaw = symDumpRaw.Replace("\"name\": \".", "\"name\": \"_");

            SymDumpSymbolCollection symbolCollection = JsonConvert.DeserializeObject<SymDumpSymbolCollection>(symDumpRaw);

            BenLincoln.DebugSymbols.FunctionSourceTracker fsTracker = new BenLincoln.DebugSymbols.FunctionSourceTracker();

            foreach (CFunction func in symbolCollection.functions)
            {
                try
                {
                    fsTracker.AddOrUpdateEntry(func);
                }
                // disabled for debugging by using an exception that's never thrown
                catch (OutOfMemoryException ex)
                {
                    Console.WriteLine(String.Format("Error parsing method '{0}', signature '{1}': {2}", func.name, func.method_signature, ex.Message));
                }
            }

            List<CFunction> decompiledFunctions = BenLincoln.DebugSymbols.SourceCodeParser.GetAllFunctionsFromSource(decompiledSource);
            List<Guid> matchedFunctionGUIDs = new List<Guid>();

            // create and populate individual source files
            foreach (string sourceFileName in fsTracker.SourceFiles.Keys)
            {
                string stubSourceFilePath = sourceFileName;
                // remove any leading drive letter if the code was compiled on Windows
                Regex driveLetterRex = new Regex(@"^[A-Za-z]:");
                if (driveLetterRex.IsMatch(stubSourceFilePath))
                {
                    stubSourceFilePath = stubSourceFilePath.Substring(2);
                }
                // replace forward and backslashes with the path separator for the current platform
                stubSourceFilePath = stubSourceFilePath.Replace("/", Path.DirectorySeparatorChar.ToString());
                stubSourceFilePath = stubSourceFilePath.Replace(@"\", Path.DirectorySeparatorChar.ToString());
                // make path relative to output directory
                while (stubSourceFilePath.StartsWith(Path.DirectorySeparatorChar.ToString()))
                {
                    stubSourceFilePath = stubSourceFilePath.Substring(1);
                }
                string stubTemp = stubSourceFilePath;
                stubSourceFilePath = Path.Combine(baseOutputDirectoryPath, "source-decompiled", stubTemp);

                // get a list of unique line numbers for methods from the file, and sort by that
                List<uint> lineNumbers = new List<uint>();
                foreach (CFunction f in fsTracker.SourceFiles[sourceFileName])
                {
                    if (!lineNumbers.Contains(f.source_file_line_number))
                    {
                        lineNumbers.Add(f.source_file_line_number);
                    }
                }
                lineNumbers.Sort();

                StringBuilder headerFileBuilder = new StringBuilder();
                StringBuilder sourceFileStubBuilder = new StringBuilder();

                string headerFilePath = Path.ChangeExtension(stubSourceFilePath, "H");
                string headerFileName = Path.GetFileName(headerFilePath);
                string headerFileDefString = headerFileName.Replace(".", "_");

                bool headerAndSourceAreSameFile = false;
                if (Path.GetFileName(stubSourceFilePath) == headerFileName)
                {
                    headerAndSourceAreSameFile = true;
                }

                headerFileBuilder.Append(String.Format("#ifndef {0}", headerFileDefString));
                headerFileBuilder.Append(Environment.NewLine);
                headerFileBuilder.Append(String.Format("#define {0}", headerFileDefString));
                headerFileBuilder.Append(Environment.NewLine);
                headerFileBuilder.Append(Environment.NewLine);

                // handle edge case: source file IS header file
                // i.e. if functions are defined in a .H file instead of just declared in it.
                if (!headerAndSourceAreSameFile)
                {
                    sourceFileStubBuilder.Append(String.Format("#include \"{0}.H\"", projectName));
                    sourceFileStubBuilder.Append(Environment.NewLine);
                    sourceFileStubBuilder.Append(String.Format("#include \"{0}\"", headerFileName));
                    sourceFileStubBuilder.Append(Environment.NewLine);
                }
                sourceFileStubBuilder.Append(Environment.NewLine);

                foreach (int lineNumber in lineNumbers)
                {
                    foreach (CFunction f in fsTracker.SourceFiles[sourceFileName])
                    {
                        if (f.source_file_line_number == lineNumber)
                        {
                            // get decompiled functions that match this one to some extent
                            List<DecompiledSourceCodeMatch> matches = new List<DecompiledSourceCodeMatch>();
                            float highestMatchScore = 0.0f;
                            foreach (CFunction otherFunc in decompiledFunctions)
                            {
                                //if (f.name == otherFunc.name)
                                //{
                                //    Console.WriteLine("Debug");
                                //}
                                if (f.offset == otherFunc.offset)
                                {
                                    DecompiledSourceCodeMatch matchingFunc = new DecompiledSourceCodeMatch();
                                    matchingFunc.DecompiledFunction = otherFunc;
                                    matchingFunc.MatchPercent = 100.0f;
                                    highestMatchScore = 100.0f;
                                    matches.Add(matchingFunc);
                                    break;
                                }
                            }
                            //foreach (CFunction otherFunc in decompiledFunctions)
                            //{
                            //    if (!matchedFunctionGUIDs.Contains(otherFunc.guid))
                            //    {
                            //        float mp = f.MatchPercentWithOtherFunction(otherFunc, true);
                            //        if (mp > 0.0f)
                            //        {
                            //            DecompiledSourceCodeMatch matchingFunc = new DecompiledSourceCodeMatch();
                            //            matchingFunc.DecompiledFunction = otherFunc;
                            //            matchingFunc.MatchPercent = mp;
                            //            matches.Add(matchingFunc);
                            //            if (mp > highestMatchScore)
                            //            {
                            //                highestMatchScore = mp;
                            //            }
                            //        }
                            //    }
                            //}
                            string functionCode = "";
                            if (matches.Count > 0)
                            {
                                DecompiledSourceCodeMatch selectedFunc = matches[0];
                                if (matches.Count > 1)
                                {
                                    bool foundBestMatch = false;
                                    if (matches[0].MatchPercent == highestMatchScore)
                                    {
                                        foundBestMatch = true;
                                    }
                                    bool multipleBestMatches = false;
                                    for (int i = 1; i < matches.Count; i++)
                                    {
                                        if (matches[i].MatchPercent == highestMatchScore)
                                        {
                                            if (!foundBestMatch)
                                            {
                                                selectedFunc = matches[i];
                                                foundBestMatch = true;
                                            }
                                            else
                                            {
                                                multipleBestMatches = true;
                                            }
                                            break;
                                        }
                                    }
                                    if (multipleBestMatches)
                                    {
                                        Console.WriteLine(string.Format("Warning: {0} matches when searching for decompiled function '{1}'. The best match was {2:F2} percent, and there were multiple matches with this score. You should manually verify the result.", matches.Count, f.name, highestMatchScore));
                                    }
                                    else
                                    {
                                        Console.WriteLine(string.Format("Warning: {0} matches when searching for decompiled function '{1}'. The best match was {2:F2} percent. You should manually verify the result.", matches.Count, f.name, highestMatchScore));
                                    }
                                }
                                else
                                {
                                    Console.WriteLine(string.Format("Found 1 match for function '{0}', with a score of {1:F2} percent.", f.name, highestMatchScore));
                                }
                                matchedFunctionGUIDs.Add(selectedFunc.DecompiledFunction.guid);
                                StringBuilder functionCodeBuilder = new StringBuilder();
                                functionCodeBuilder.Append(matches[0].DecompiledFunction.GetFullCode());
                                functionCode = functionCodeBuilder.ToString();
                            }
                            else
                            {
                                Console.WriteLine(string.Format("Did not find any matches for function '{0}' in the decompiled code. A stub will be created instead.", f.name));
                            }

                            headerFileBuilder.Append(f.GetHeaderDefinition());
                            headerFileBuilder.Append(Environment.NewLine);
                            headerFileBuilder.Append(Environment.NewLine);

                            sourceFileStubBuilder.Append(f.GetFunctionStub(functionCode));
                            sourceFileStubBuilder.Append(Environment.NewLine);
                            sourceFileStubBuilder.Append(Environment.NewLine);
                        }
                    }
                }

                //sourceFileStubBuilder.Append("/* Begin functions not in symbol tables */");
                //sourceFileStubBuilder.Append(Environment.NewLine);
                //sourceFileStubBuilder.Append(Environment.NewLine);

                //foreach (CFunction otherFunc in decompiledFunctions)
                //{
                //    if (!matchedFunctionGUIDs.Contains(otherFunc.guid))
                //    {
                //        sourceFileStubBuilder.Append(otherFunc.GetFunctionStub(otherFunc.GetFullCode()));
                //        sourceFileStubBuilder.Append(Environment.NewLine);
                //        sourceFileStubBuilder.Append(Environment.NewLine);
                //    }
                //}

                //sourceFileStubBuilder.Append("/* End functions not in symbol tables */");
                //sourceFileStubBuilder.Append(Environment.NewLine);
                //sourceFileStubBuilder.Append(Environment.NewLine);
                if (headerAndSourceAreSameFile)
                {
                    headerFileBuilder.Append("%STUBCONTENT%");
                    headerFileBuilder.Append(Environment.NewLine);
                }

                headerFileBuilder.Append(String.Format("#endif", headerFileDefString));

                string parentPath = Path.GetDirectoryName(stubSourceFilePath);
                if (!Directory.Exists(parentPath))
                {
                    try
                    {
                        Directory.CreateDirectory(parentPath);
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(String.Format("Could not create a required output path ('{0}'): {1}", parentPath, ex.Message));
                        Environment.Exit(5);
                    }
                }
                string headerFileContent = headerFileBuilder.ToString();
                string sourceFileStubContent = sourceFileStubBuilder.ToString();
                if (headerAndSourceAreSameFile)
                {
                    headerFileContent = headerFileContent.Replace("%STUBCONTENT%", sourceFileStubContent);
                    WriteToOutput(headerFilePath, headerFileContent, FileMode.Create);
                }
                else
                {
                    WriteToOutput(headerFilePath, headerFileContent, FileMode.Create);
                    WriteToOutput(stubSourceFilePath, sourceFileStubContent, FileMode.Create);
                }
            }

            // create a file with all of the unmatched functions from the decompiled source
            string unmatchedFunctionSourceFilePath = Path.Combine(baseOutputDirectoryPath, "source-decompiled", "Unmatched_Decompiled_Functions.C");
            StringBuilder unmatchedBuilder = new StringBuilder();
            for (int i = 0; i < decompiledFunctions.Count; i++)
            {
                if (!matchedFunctionGUIDs.Contains(decompiledFunctions[i].guid))
                {
                    manualChangesBuilder.Append(string.Format("The decompiled function '{0}' was not matched to a source file, and has been placed in '{1}'", decompiledFunctions[i].name, unmatchedFunctionSourceFilePath));
                    manualChangesBuilder.Append(Environment.NewLine);
                    //unmatchedBuilder.Append("// ");
                    //unmatchedBuilder.Append(decompiledFunctions[i].offset);
                    //unmatchedBuilder.Append(Environment.NewLine);
                    //unmatchedBuilder.Append(decompiledFunctions[i].GetFullCode());
                    unmatchedBuilder.Append(decompiledFunctions[i].GetFunctionStub(decompiledFunctions[i].GetFullCode()));
                    unmatchedBuilder.Append(Environment.NewLine);
                    unmatchedBuilder.Append(Environment.NewLine);
                    unmatchedBuilder.Append(Environment.NewLine);
                }
            }

            WriteToOutput(unmatchedFunctionSourceFilePath, unmatchedBuilder.ToString(), FileMode.Create);

            // create a list of known manual actions that need to be taken by the user
            manualChangesBuilder.Append(Environment.NewLine);
            WriteToOutput(manualChangesPath, manualChangesBuilder.ToString(), FileMode.Create);

            Console.WriteLine("");
        }

        static void DisplayHelpMessage()
        {
            DisplayBanner();

            Console.Write(GetEmbeddedTextFile("CreateSkeleton.HelpMessage.txt"));
            Console.WriteLine();
        }

        protected static string GetEmbeddedTextFile(string fileName)
        {
            Assembly asm = Assembly.GetExecutingAssembly();
            Stream inStream = asm.GetManifestResourceStream(fileName);
            StreamReader inReader = new StreamReader(inStream);
            return inReader.ReadToEnd();
        }

        static void DisplayBanner()
        {
            Assembly currentAssembly = Assembly.GetExecutingAssembly();

            AssemblyName assemblyName = currentAssembly.GetName();

            int versionMajor = assemblyName.Version.Major;
            int versionMinor = assemblyName.Version.Minor;

            Console.WriteLine(string.Format("{0} {1}.{2}", ProgramName, versionMajor, versionMinor));
            Console.WriteLine(Copyright);
            Console.WriteLine(InfoURL);
            Console.WriteLine(Environment.NewLine);
        }

        protected static void ExitWithError(string message, int exitCode)
        {
            DisplayBanner();
            Console.WriteLine(message);
            System.Environment.ExitCode = exitCode;
            Environment.Exit(exitCode);
        }

        private static void WriteToOutput(string outputPath, string text, FileMode mode)
        {
            FileStream oStream;
            StreamWriter oWriter;

            oStream = new FileStream(outputPath, mode, FileAccess.Write);
            oWriter = new StreamWriter(oStream);

            oWriter.WriteLine(text);

            oWriter.Close();
            oStream.Close();
        }

        private static string GetFileAsString(string inputPath)
        {
            string result = "";
            FileStream inStream;
            StreamReader inReader;

            inStream = new FileStream(inputPath, FileMode.Open, FileAccess.Read);
            inReader = new StreamReader(inStream);

            result = inReader.ReadToEnd();

            inReader.Close();
            inStream.Close();
            return result;
        }

        private static string[] GetLinesAsStringArray(string inputString)
        {
            // for easier cross-platform use, this handles Unix and DOS-style newlines the same way
            string inString = inputString.Replace("\r", "");

            string[] result = inString.Split(new string[] { "\n" }, StringSplitOptions.RemoveEmptyEntries);

            return result;
        }

        private static string[] GetFileLinesAsStringArray(string inputPath)
        {
            return GetLinesAsStringArray(GetFileAsString(inputPath));
        }
    }
}
